"""Various high level ways to transform a complete Form into a new Form."""
# Copyright (C) 2008-2016 Martin Sandve Alnæs
#
# This file is part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
#
# Modified by Anders Logg, 2009
# Modified by Massimiliano Leoni, 2016
# Modified by Cecile Daversin-Catty, 2018

from ufl.action import Action
from ufl.adjoint import Adjoint
from ufl.algorithms import (
    compute_energy_norm,
    compute_form_action,
    compute_form_adjoint,
    compute_form_functional,
    compute_form_lhs,
    compute_form_rhs,
    expand_derivatives,
    extract_arguments,
    formsplitter,
    replace,  # noqa: F401
)
from ufl.argument import Argument
from ufl.coefficient import Coefficient, Cofunction
from ufl.constantvalue import as_ufl, is_true_ufl_scalar
from ufl.core.base_form_operator import BaseFormOperator
from ufl.core.expr import Expr, ufl_err_str
from ufl.core.multiindex import FixedIndex, MultiIndex
from ufl.differentiation import (
    BaseFormCoordinateDerivative,
    BaseFormDerivative,
    BaseFormOperatorCoordinateDerivative,
    BaseFormOperatorDerivative,
    CoefficientDerivative,
    CoordinateDerivative,
)
from ufl.exprcontainers import ExprList, ExprMapping
from ufl.finiteelement import MixedElement
from ufl.form import BaseForm, Form, FormSum, ZeroBaseForm, as_form
from ufl.functionspace import FunctionSpace
from ufl.geometry import SpatialCoordinate
from ufl.indexed import Indexed
from ufl.sorting import sorted_expr
from ufl.split_functions import split
from ufl.tensors import ListTensor, as_tensor
from ufl.variable import Variable


def extract_blocks(form, i=None, j=None):
    """Extract blocks.

    Given a linear or bilinear form on a mixed space,
    extract the block corresponding to the indices ix, iy.

    Example:
       a = inner(grad(u), grad(v))*dx + div(u)*q*dx + div(v)*p*dx
       extract_blocks(a, 0, 0) -> inner(grad(u), grad(v))*dx
       extract_blocks(a) -> [inner(grad(u), grad(v))*dx, div(v)*p*dx, div(u)*q*dx, 0]
    """
    return formsplitter.extract_blocks(form, i, j)


def lhs(form):
    """Get the left hand side.

    Given a combined bilinear and linear form,
    extract the left hand side (bilinear form part).

    Example:
        a = u*v*dx + f*v*dx
        a = lhs(a) -> u*v*dx
    """
    form = as_form(form)
    form = expand_derivatives(form)
    return compute_form_lhs(form)


def rhs(form):
    """Get the right hand side.

    Given a combined bilinear and linear form,
    extract the right hand side (negated linear form part).

    Example:
        a = u*v*dx + f*v*dx
        L = rhs(a) -> -f*v*dx
    """
    form = as_form(form)
    form = expand_derivatives(form)
    return compute_form_rhs(form)


def system(form):
    """Split a form into the left hand side and right hand side.

    See ``lhs`` and ``rhs``.
    """
    return lhs(form), rhs(form)


def functional(form):  # TODO: Does this make sense for anything other than testing?
    """Extract the functional part of form."""
    form = as_form(form)
    form = expand_derivatives(form)
    return compute_form_functional(form)


def action(form, coefficient=None, derivatives_expanded=None):
    """Get the action.

    Given a bilinear form, return a linear form
    with an additional coefficient, representing the
    action of the form on the coefficient. This can be
    used for matrix-free methods.
    For formbase objects,coefficient can be any object of the correct type,
    and this function returns an Action object.

    When `action` is being called multiple times on the same form, expanding derivatives
    become expensive -> `derivatives_expanded` enables to use caching mechanisms to avoid that.
    """
    form = as_form(form)
    is_coefficient_valid = not isinstance(coefficient, BaseForm) or (
        isinstance(coefficient, BaseFormOperator) and len(coefficient.arguments()) == 1
    )
    # Can't expand derivatives on objects that are not Form or Expr (e.g. Matrix)
    if isinstance(form, (Form, BaseFormOperator)) and is_coefficient_valid:
        if not derivatives_expanded:
            # For external operators differentiation may turn a Form into a FormSum
            form = expand_derivatives(form)
        if isinstance(form, Form):
            return compute_form_action(form, coefficient)
    return Action(form, coefficient)


def energy_norm(form, coefficient=None):
    """Get the energy norm.

    Given a bilinear form *a* and a coefficient *f*,
    return the functional :math:`a(f,f)`.
    """
    form = as_form(form)
    form = expand_derivatives(form)
    return compute_energy_norm(form, coefficient)


def adjoint(form, reordered_arguments=None, derivatives_expanded=None):
    """Get the adjoint.

    Given a combined bilinear form, compute the adjoint form by
    changing the ordering (count) of the test and trial functions, and
    taking the complex conjugate of the result.

    By default, new ``Argument`` objects will be created with
    opposite ordering. However, if the adjoint form is to
    be added to other forms later, their arguments must match.
    In that case, the user must provide a tuple *reordered_arguments*=(u2,v2).

    If the form is a baseform instance instead of a Form object, we return an Adjoint
    object instructing the adjoint to be computed at a later point.

    When `adjoint` is being called multiple times on the same form, expanding derivatives
    become expensive -> `derivatives_expanded` enables to use caching mechanisms to avoid that.
    """
    form = as_form(form)
    if isinstance(form, BaseForm):
        # Allow BaseForm objects that are not BaseForm such as Adjoint since there are cases
        # where we need to expand derivatives: e.g. to get the number of arguments
        #   => For example: Adjoint(Action(2-form, derivative(u,u)))
        if not derivatives_expanded:
            # For external operators differentiation may turn a Form into a FormSum
            form = expand_derivatives(form)
        if isinstance(form, Form):
            return compute_form_adjoint(form, reordered_arguments)
    return Adjoint(form)


def zero_lists(shape):
    """Createa list of zeros of the given shape."""
    if len(shape) == 0:
        raise ValueError("Invalid shape.")
    elif len(shape) == 1:
        return [0] * shape[0]
    else:
        return [zero_lists(shape[1:]) for i in range(shape[0])]


def set_list_item(li, i, v):
    """Set an item in a nested list."""
    # Get to the innermost list
    if len(i) > 1:
        for j in i[:-1]:
            li = li[j]
    # Set item in innermost list
    li[i[-1]] = v


def _handle_derivative_arguments(form, coefficient, argument):
    """Handle derivative arguments."""
    # Wrap single coefficient in tuple for uniform treatment below
    if isinstance(coefficient, (list, tuple, ListTensor)):
        coefficients = tuple(coefficient)
    else:
        coefficients = (coefficient,)

    if argument is None:
        # Try to create argument if not provided
        if not all(
            isinstance(c, (Coefficient, Cofunction, BaseFormOperator)) for c in coefficients
        ):
            raise ValueError(
                "Can only create arguments automatically for non-indexed coefficients."
            )

        # Get existing arguments from form and position the new one
        # with the next argument number
        if isinstance(form, Form):
            form_arguments = form.arguments()
        else:
            # To handle derivative(expression), which is at least used
            # in tests. Remove?
            form_arguments = extract_arguments(form)

        numbers = sorted(set(arg.number() for arg in form_arguments))
        number = max(numbers + [-1]) + 1

        # Don't know what to do with parts, let the user sort it out
        # in that case
        parts = set(arg.part() for arg in form_arguments)
        if len(parts - {None}) != 0:
            raise ValueError("Not expecting parts here, provide your own arguments.")
        part = None

        # Create argument and split it if in a mixed space
        function_spaces = [c.ufl_function_space() for c in coefficients]
        if len(function_spaces) == 1:
            arguments = (Argument(function_spaces[0], number, part),)
        else:
            # Create in mixed space over assumed (for now) same domain
            domains = [fs.ufl_domain() for fs in function_spaces]
            elements = [fs.ufl_element() for fs in function_spaces]
            assert all(fs.ufl_domain() == domains[0] for fs in function_spaces)
            elm = MixedElement(elements)
            fs = FunctionSpace(domains[0], elm)
            arguments = split(Argument(fs, number, part))
    else:
        # Wrap single argument in tuple for uniform treatment below
        if isinstance(argument, (list, tuple)):
            arguments = tuple(argument)
        else:
            n = len(coefficients)
            if n == 1:
                arguments = (argument,)
            else:
                if argument.ufl_shape == (n,):
                    arguments = tuple(argument[i] for i in range(n))
                else:
                    arguments = split(argument)

    # Build mapping from coefficient to argument
    m = {}
    for c, a in zip(coefficients, arguments):
        if c.ufl_shape != a.ufl_shape:
            raise ValueError("Coefficient and argument shapes do not match!")
        if isinstance(c, (Coefficient, Cofunction, BaseFormOperator, SpatialCoordinate)):
            m[c] = a
        else:
            if not isinstance(c, Indexed):
                raise ValueError(f"Invalid coefficient type for {ufl_err_str(c)}")
            f, i = c.ufl_operands
            if not isinstance(f, Coefficient):
                raise ValueError(f"Expecting an indexed coefficient, not {ufl_err_str(f)}")
            if not (isinstance(i, MultiIndex) and all(isinstance(j, FixedIndex) for j in i)):
                raise ValueError(f"Expecting one or more fixed indices, not {ufl_err_str(i)}")
            i = tuple(int(j) for j in i)
            if f not in m:
                m[f] = {}
            m[f][i] = a

    # Merge coefficient derivatives (arguments) based on indices
    for c, p in m.items():
        if isinstance(p, dict):
            a = zero_lists(c.ufl_shape)
            for i, g in p.items():
                set_list_item(a, i, g)
            m[c] = as_tensor(a)

    # Wrap and return generic tuples
    items = sorted(m.items(), key=lambda x: x[0].count())
    coefficients = ExprList(*[item[0] for item in items])
    arguments = ExprList(*[item[1] for item in items])
    return coefficients, arguments


def derivative(form, coefficient, argument=None, coefficient_derivatives=None):
    """Compute the Gateaux derivative of *form* w.r.t. *coefficient* in direction of *argument*.

    If the argument is omitted, a new ``Argument`` is created
    in the same space as the coefficient, with argument number
    one higher than the highest one in the form.

    The resulting form has one additional ``Argument``
    in the same finite element space as the coefficient.

    A tuple of ``Coefficient`` s may be provided in place of
    a single ``Coefficient``, in which case the new ``Argument``
    argument is based on a ``MixedElement`` created from this tuple.

    An indexed ``Coefficient`` from a mixed space may be provided,
    in which case the argument should be in the corresponding
    subspace of the coefficient space.

    If provided, *coefficient_derivatives* should be a mapping from
    ``Coefficient`` instances to their derivatives w.r.t. *coefficient*.
    """
    if isinstance(form, FormSum):
        # Distribute derivative over FormSum components
        return FormSum(
            *[
                (derivative(component, coefficient, argument, coefficient_derivatives), 1)
                for component in form.components()
            ]
        )
    elif isinstance(form, Adjoint):
        # Is `derivative(Adjoint(A), ...)` with A a 2-form even legal ?
        # -> If yes, what's the right thing to do here ?
        raise NotImplementedError("Adjoint derivative is not supported.")
    elif isinstance(form, Action):
        # Push derivative through Action slots
        left, right = form.ufl_operands
        # Eagerly simplify spatial derivatives when Action results in a scalar.
        if not len(form.arguments()) and isinstance(coefficient, SpatialCoordinate):
            return ZeroBaseForm(())

        if len(left.arguments()) == 1:
            dleft = derivative(left, coefficient, argument, coefficient_derivatives)
            dright = derivative(right, coefficient, argument, coefficient_derivatives)
            # Leibniz formula
            return action(
                adjoint(dleft, derivatives_expanded=True), right, derivatives_expanded=True
            ) + action(left, dright, derivatives_expanded=True)
        else:
            raise NotImplementedError(
                "Action derivative not supported when the left argument is not a 1-form."
            )

    coefficients, arguments = _handle_derivative_arguments(form, coefficient, argument)
    if coefficient_derivatives is None:
        coefficient_derivatives = ExprMapping()
    else:
        cd = []
        for k in sorted_expr(coefficient_derivatives.keys()):
            cd += [as_ufl(k), as_ufl(coefficient_derivatives[k])]
        coefficient_derivatives = ExprMapping(*cd)

    # Got a form? Apply derivatives to the integrands in turn.
    if isinstance(form, Form):
        integrals = []
        for itg in form.integrals():
            if isinstance(coefficient, SpatialCoordinate):
                fd = CoordinateDerivative(
                    itg.integrand(), coefficients, arguments, coefficient_derivatives
                )
            elif isinstance(coefficient, BaseForm) and not isinstance(
                coefficient, BaseFormOperator
            ):
                # Make the `ZeroBaseForm` arguments
                arguments = form.arguments() + coefficient.arguments()
                return ZeroBaseForm(arguments)
            else:
                fd = CoefficientDerivative(
                    itg.integrand(), coefficients, arguments, coefficient_derivatives
                )
            integrals.append(itg.reconstruct(fd))
        return Form(integrals)

    elif isinstance(form, BaseFormOperator):
        if not isinstance(coefficient, SpatialCoordinate):
            return BaseFormOperatorDerivative(
                form, coefficients, arguments, coefficient_derivatives
            )
        else:
            return BaseFormOperatorCoordinateDerivative(
                form, coefficients, arguments, coefficient_derivatives
            )

    elif isinstance(form, BaseForm):
        if not isinstance(coefficient, SpatialCoordinate):
            return BaseFormDerivative(form, coefficients, arguments, coefficient_derivatives)
        else:
            return BaseFormCoordinateDerivative(
                form, coefficients, arguments, coefficient_derivatives
            )

    elif isinstance(form, Expr):
        # What we got was in fact an integrand
        if not isinstance(coefficient, SpatialCoordinate):
            return CoefficientDerivative(form, coefficients, arguments, coefficient_derivatives)
        else:
            return CoordinateDerivative(form, coefficients, arguments, coefficient_derivatives)

    raise ValueError(f"Invalid argument type {type(form)}.")


def sensitivity_rhs(a, u, L, v):
    r"""Compute the right hand side for a sensitivity calculation system.

    The derivation behind this computation is as follows.
    Assume *a*, *L* to be bilinear and linear forms
    corresponding to the assembled linear system

    .. math::

        Ax = b.

    Where *x* is the vector of the discrete function corresponding to *u*.
    Let *v* be some scalar variable this equation depends on.
    Then we can write

    .. math::
        0 = \\frac{d}{dv}(Ax-b) = \\frac{dA}{dv} x + A \\frac{dx}{dv} -
        \\frac{db}{dv},

        A \\frac{dx}{dv} = \\frac{db}{dv} - \\frac{dA}{dv} x,

    and solve this system for :math:`\\frac{dx}{dv}`, using the same bilinear
    form *a* and matrix *A* from the original system.
    Assume the forms are written
    ::

        v = variable(v_expression)
        L = IL(v) * dx
        a = Ia(v) * dx

    where ``IL`` and ``Ia`` are integrand expressions.
    Define a ``Coefficient u`` representing the solution
    to the equations. Then we can compute :math:`\\frac{db}{dv}`
    and :math:`\\frac{dA}{dv}` from the forms
    ::

        da = diff(a, v)
        dL = diff(L, v)

    and the action of ``da`` on ``u`` by
    ::

        dau = action(da, u)

    In total, we can build the right hand side of the system
    to compute :math:`\\frac{du}{dv}` with the single line
    ::

        dL = diff(L, v) - action(diff(a, v), u)

    or, using this function,
    ::

        dL = sensitivity_rhs(a, u, L, v)
    """
    if not (
        isinstance(a, Form)
        and isinstance(u, Coefficient)
        and isinstance(L, Form)
        and isinstance(v, Variable)
    ):
        raise ValueError(
            "Expecting (a, u, L, v), (bilinear form, function, linear form and scalar variable)."
        )
    if not is_true_ufl_scalar(v):
        raise ValueError("Expecting scalar variable.")
    from ufl.operators import diff

    return diff(L, v) - action(diff(a, v), u)



"""Algorithm for replacing gradients in an expression."""

# Copyright (C) 2013-2016 Martin Sandve Alnæs
#
# This file is part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later

from ufl.algorithms.apply_geometry_lowering import apply_geometry_lowering
from ufl.algorithms.estimate_degrees import estimate_total_polynomial_degree
from ufl.classes import (
    FacetJacobianDeterminant,
    Form,
    Integral,
    JacobianDeterminant,
    QuadratureWeight,
)
from ufl.differentiation import CoordinateDerivative
from ufl.measure import custom_integral_types, point_integral_types


def compute_integrand_scaling_factor(integral):
    """Change integrand geometry to the right representations."""
    domain = integral.ufl_domain()
    integral_type = integral.integral_type()
    # co = CellOrientation(domain)
    weight = QuadratureWeight(domain)
    tdim = domain.topological_dimension()
    # gdim = domain.geometric_dimension()

    # Polynomial degree of integrand scaling
    degree = 0
    if integral_type == "cell":
        if tdim > 0:
            detJ = JacobianDeterminant(domain)
            degree = estimate_total_polynomial_degree(apply_geometry_lowering(detJ))
            # Despite the abs, |detJ| is polynomial except for
            # self-intersecting cells, where we have other problems.
            scale = abs(detJ) * weight
        else:
            # No need to scale 'integral' over a vertex
            scale = 1

    elif integral_type.startswith("exterior_facet"):
        if tdim > 1:
            # Scaling integral by facet jacobian determinant and
            # quadrature weight
            detFJ = FacetJacobianDeterminant(domain)
            degree = estimate_total_polynomial_degree(apply_geometry_lowering(detFJ))
            scale = detFJ * weight
        else:
            # No need to scale 'integral' over a vertex
            scale = 1

    elif integral_type.startswith("interior_facet"):
        if tdim > 1:
            # Scaling integral by facet jacobian determinant from one
            # side and quadrature weight
            detFJ = FacetJacobianDeterminant(domain)
            degree = estimate_total_polynomial_degree(apply_geometry_lowering(detFJ))
            scale = detFJ("+") * weight
        else:
            # No need to scale 'integral' over a vertex
            scale = 1

    elif integral_type in custom_integral_types:
        # Scaling with custom weight, which includes eventual volume
        # scaling
        scale = weight

    elif integral_type in point_integral_types:
        # No need to scale 'integral' over a point
        scale = 1

    else:
        raise ValueError(f"Unknown integral type {integral_type}, don't know how to scale.")

    return scale, degree


def apply_integral_scaling(form):
    """Multiply integrands by a factor to scale the integral to reference frame."""
    # TODO: Consider adding an in_reference_frame property to Integral
    #       and checking it here and setting it in the returned form
    if isinstance(form, Form):
        newintegrals = [apply_integral_scaling(integral) for integral in form.integrals()]
        return Form(newintegrals)

    elif isinstance(form, Integral):
        integral = form
        integrand = integral.integrand()
        # Compute and apply integration scaling factor since we want to compute
        # coordinate derivatives at the end, the scaling factor has to be moved
        # inside those
        scale, degree = compute_integrand_scaling_factor(integral)
        md = {}
        md.update(integral.metadata())
        new_degree = degree
        cur_degree = md.get("estimated_polynomial_degree")
        if cur_degree is not None:
            if isinstance(cur_degree, tuple) and isinstance(degree, tuple):
                new_degree = tuple(d[0] + d[1] for d in zip(cur_degree, degree))
            elif isinstance(cur_degree, tuple):
                new_degree = tuple(d + degree for d in cur_degree)
            elif isinstance(degree, tuple):
                new_degree = tuple(cur_degree + d for d in degree)
            else:
                new_degree = cur_degree + degree
        md["estimated_polynomial_degree"] = new_degree

        def scale_coordinate_derivative(o, scale):
            """Scale the coordinate derivative."""
            o_ = o.ufl_operands
            if isinstance(o, CoordinateDerivative):
                return CoordinateDerivative(
                    scale_coordinate_derivative(o_[0], scale), o_[1], o_[2], o_[3]
                )
            else:
                return scale * o

        newintegrand = scale_coordinate_derivative(integrand, scale)
        return integral.reconstruct(integrand=newintegrand, metadata=md)

    else:
        raise ValueError(f"Invalid type {form.__class__.__name__}")



"""Algorithm for lowering abstractions of geometric types.

This means replacing high-level types with expressions
of mostly the Jacobian and reference cell data.
"""

# Copyright (C) 2013-2016 Martin Sandve Alnæs
#
# This file is part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later

import warnings
from functools import reduce
from itertools import combinations

from ufl.classes import (
    CellCoordinate,
    CellEdgeVectors,
    CellFacetJacobian,
    CellOrientation,
    CellOrigin,
    CellVertices,
    CellVolume,
    Expr,
    FacetEdgeVectors,
    FacetJacobian,
    FacetJacobianDeterminant,
    FloatValue,
    Form,
    Integral,
    Jacobian,
    JacobianDeterminant,
    JacobianInverse,
    MaxCellEdgeLength,
    ReferenceCellVolume,
    ReferenceFacetVolume,
    ReferenceGrad,
    ReferenceNormal,
    SpatialCoordinate,
)
from ufl.compound_expressions import cross_expr, determinant_expr, inverse_expr
from ufl.core.multiindex import Index, indices
from ufl.corealg.map_dag import map_expr_dag
from ufl.corealg.multifunction import MultiFunction, memoized_handler
from ufl.domain import extract_unique_domain
from ufl.measure import custom_integral_types, point_integral_types
from ufl.operators import conj, max_value, min_value, real, sqrt
from ufl.tensors import as_tensor, as_vector


class GeometryLoweringApplier(MultiFunction):
    """Geometry lowering."""

    def __init__(self, preserve_types=()):
        """Initialise."""
        MultiFunction.__init__(self)
        # Store preserve_types as boolean lookup table
        self._preserve_types = [False] * Expr._ufl_num_typecodes_
        for cls in preserve_types:
            self._preserve_types[cls._ufl_typecode_] = True

    expr = MultiFunction.reuse_if_untouched

    def terminal(self, t):
        """Apply to terminal."""
        return t

    @memoized_handler
    def jacobian(self, o):
        """Apply to jacobian."""
        if self._preserve_types[o._ufl_typecode_]:
            return o
        domain = extract_unique_domain(o)
        if not domain.ufl_coordinate_element().pullback.is_identity:
            raise ValueError("Piola mapped coordinates are not implemented.")
        # Note: No longer supporting domain.coordinates(), always
        # preserving SpatialCoordinate object.  However if Jacobians
        # are not preserved, using
        # ReferenceGrad(SpatialCoordinate(domain)) to represent them.
        x = self.spatial_coordinate(SpatialCoordinate(domain))
        return ReferenceGrad(x)

    @memoized_handler
    def _future_jacobian(self, o):
        """Apply to _future_jacobian."""
        # If we're not using Coefficient to represent the spatial
        # coordinate, we can just as well just return o here too
        # unless we add representation of basis functions and dofs to
        # the ufl layer (which is nice to avoid).
        return o

    @memoized_handler
    def jacobian_inverse(self, o):
        """Apply to jacobian_inverse."""
        if self._preserve_types[o._ufl_typecode_]:
            return o

        domain = extract_unique_domain(o)
        J = self.jacobian(Jacobian(domain))
        # TODO: This could in principle use
        # preserve_types[JacobianDeterminant] with minor refactoring:
        K = inverse_expr(J)
        return K

    @memoized_handler
    def jacobian_determinant(self, o):
        """Apply to jacobian_determinant."""
        if self._preserve_types[o._ufl_typecode_]:
            return o

        domain = extract_unique_domain(o)
        J = self.jacobian(Jacobian(domain))
        detJ = determinant_expr(J)

        # TODO: Is "signing" the determinant for manifolds the
        #       cleanest approach?  The alternative is to have a
        #       specific type for the unsigned pseudo-determinant.
        if domain.topological_dimension() < domain.geometric_dimension():
            co = CellOrientation(domain)
            detJ = co * detJ

        return detJ

    @memoized_handler
    def facet_jacobian(self, o):
        """Apply to facet_jacobian."""
        if self._preserve_types[o._ufl_typecode_]:
            return o

        domain = extract_unique_domain(o)
        J = self.jacobian(Jacobian(domain))
        RFJ = CellFacetJacobian(domain)
        i, j, k = indices(3)
        return as_tensor(J[i, k] * RFJ[k, j], (i, j))

    @memoized_handler
    def facet_jacobian_inverse(self, o):
        """Apply to facet_jacobian_inverse."""
        if self._preserve_types[o._ufl_typecode_]:
            return o

        domain = extract_unique_domain(o)
        FJ = self.facet_jacobian(FacetJacobian(domain))
        # This could in principle use
        # preserve_types[JacobianDeterminant] with minor refactoring:
        return inverse_expr(FJ)

    @memoized_handler
    def facet_jacobian_determinant(self, o):
        """Apply to facet_jacobian_determinant."""
        if self._preserve_types[o._ufl_typecode_]:
            return o

        domain = extract_unique_domain(o)
        FJ = self.facet_jacobian(FacetJacobian(domain))
        detFJ = determinant_expr(FJ)

        # TODO: Should we "sign" the facet jacobian determinant for
        #       manifolds?  It's currently used unsigned in
        #       apply_integral_scaling.
        # if domain.topological_dimension() < domain.geometric_dimension():
        #     co = CellOrientation(domain)
        #     detFJ = co*detFJ

        return detFJ

    @memoized_handler
    def spatial_coordinate(self, o):
        """Apply to spatial_coordinate.

        Fall through to coordinate field of domain if it exists.
        """
        if self._preserve_types[o._ufl_typecode_]:
            return o
        if not extract_unique_domain(o).ufl_coordinate_element().pullback.is_identity:
            raise ValueError("Piola mapped coordinates are not implemented.")
        # No longer supporting domain.coordinates(), always preserving
        # SpatialCoordinate object.
        return o

    @memoized_handler
    def cell_coordinate(self, o):
        """Apply to cell_coordinate.

        Compute from physical coordinates if they are known, using the appropriate mappings.
        """
        if self._preserve_types[o._ufl_typecode_]:
            return o

        domain = extract_unique_domain(o)
        K = self.jacobian_inverse(JacobianInverse(domain))
        x = self.spatial_coordinate(SpatialCoordinate(domain))
        x0 = CellOrigin(domain)
        i, j = indices(2)
        X = as_tensor(K[i, j] * (x[j] - x0[j]), (i,))
        return X

    @memoized_handler
    def facet_cell_coordinate(self, o):
        """Apply to facet_cell_coordinate."""
        if self._preserve_types[o._ufl_typecode_]:
            return o

        raise ValueError(
            "Missing computation of facet reference coordinates "
            "from physical coordinates via mappings."
        )

    @memoized_handler
    def cell_volume(self, o):
        """Apply to cell_volume."""
        if self._preserve_types[o._ufl_typecode_]:
            return o

        domain = extract_unique_domain(o)
        if not domain.is_piecewise_linear_simplex_domain():
            # Don't lower for non-affine cells, instead leave it to
            # form compiler
            warnings.warn("Only know how to compute the cell volume of an affine cell.")
            return o

        r = self.jacobian_determinant(JacobianDeterminant(domain))
        r0 = ReferenceCellVolume(domain)
        return abs(r * r0)

    @memoized_handler
    def facet_area(self, o):
        """Apply to facet_area."""
        if self._preserve_types[o._ufl_typecode_]:
            return o

        domain = extract_unique_domain(o)
        tdim = domain.topological_dimension()
        if not domain.is_piecewise_linear_simplex_domain():
            # Don't lower for non-affine cells, instead leave it to
            # form compiler
            warnings.warn("Only know how to compute the facet area of an affine cell.")
            return o

        # Area of "facet" of interval (i.e. "area" of a vertex) is defined as 1.0
        if tdim == 1:
            return FloatValue(1.0)

        r = self.facet_jacobian_determinant(FacetJacobianDeterminant(domain))
        r0 = ReferenceFacetVolume(domain)
        return abs(r * r0)

    @memoized_handler
    def circumradius(self, o):
        """Apply to circumradius."""
        if self._preserve_types[o._ufl_typecode_]:
            return o

        domain = extract_unique_domain(o)

        if not domain.is_piecewise_linear_simplex_domain():
            raise ValueError("Circumradius only makes sense for affine simplex cells")

        cellname = domain.ufl_cell().cellname()
        cellvolume = self.cell_volume(CellVolume(domain))

        if cellname == "interval":
            # Optimization for square interval; no square root needed
            return 0.5 * cellvolume

        # Compute lengths of cell edges
        edges = CellEdgeVectors(domain)
        num_edges = edges.ufl_shape[0]
        j = Index()
        elen = [real(sqrt(real(edges[e, j] * conj(edges[e, j])))) for e in range(num_edges)]

        if cellname == "triangle":
            return (elen[0] * elen[1] * elen[2]) / (4.0 * cellvolume)

        elif cellname == "tetrahedron":
            # la, lb, lc = lengths of the sides of an intermediate triangle
            # NOTE: Is here some hidden numbering assumption?
            la = elen[3] * elen[2]
            lb = elen[4] * elen[1]
            lc = elen[5] * elen[0]
            # p = perimeter
            p = la + lb + lc
            # s = semiperimeter
            s = p / 2
            # area of intermediate triangle with Herons formula
            triangle_area = sqrt(s * (s - la) * (s - lb) * (s - lc))
            return triangle_area / (6.0 * cellvolume)

    @memoized_handler
    def max_cell_edge_length(self, o):
        """Apply to max_cell_edge_length."""
        return self._reduce_cell_edge_length(o, max_value)

    @memoized_handler
    def min_cell_edge_length(self, o):
        """Apply to min_cell_edge_length."""
        return self._reduce_cell_edge_length(o, min_value)

    def _reduce_cell_edge_length(self, o, reduction_op):
        """Apply to _reduce_cell_edge_length."""
        if self._preserve_types[o._ufl_typecode_]:
            return o

        domain = extract_unique_domain(o)

        if domain.ufl_coordinate_element().embedded_subdegree > 1:
            # Don't lower bendy cells, instead leave it to form compiler
            warnings.warn("Only know how to compute cell edge lengths of P1 or Q1 cell.")
            return o

        elif domain.ufl_cell().cellname() == "interval":
            # Interval optimization, square root not needed
            return self.cell_volume(CellVolume(domain))

        else:
            # Other P1 or Q1 cells
            edges = CellEdgeVectors(domain)
            num_edges = edges.ufl_shape[0]
            j = Index()
            elen2 = [real(edges[e, j] * conj(edges[e, j])) for e in range(num_edges)]
            return real(sqrt(reduce(reduction_op, elen2)))

    @memoized_handler
    def cell_diameter(self, o):
        """Apply to cell_diameter."""
        if self._preserve_types[o._ufl_typecode_]:
            return o

        domain = extract_unique_domain(o)

        if domain.ufl_coordinate_element().embedded_subdegree > 1:
            # Don't lower bendy cells, instead leave it to form compiler
            warnings.warn("Only know how to compute cell diameter of P1 or Q1 cell.")
            return o

        elif domain.is_piecewise_linear_simplex_domain():
            # Simplices
            return self.max_cell_edge_length(MaxCellEdgeLength(domain))

        else:
            # Q1 cells, maximal distance between any two vertices
            verts = CellVertices(domain)
            verts = [verts[v, ...] for v in range(verts.ufl_shape[0])]
            j = Index()
            elen2 = (real((v0 - v1)[j] * conj((v0 - v1)[j])) for v0, v1 in combinations(verts, 2))
            return real(sqrt(reduce(max_value, elen2)))

    @memoized_handler
    def max_facet_edge_length(self, o):
        """Apply to max_facet_edge_length."""
        return self._reduce_facet_edge_length(o, max_value)

    @memoized_handler
    def min_facet_edge_length(self, o):
        """Apply to min_facet_edge_length."""
        return self._reduce_facet_edge_length(o, min_value)

    def _reduce_facet_edge_length(self, o, reduction_op):
        """Apply to _reduce_facet_edge_length."""
        if self._preserve_types[o._ufl_typecode_]:
            return o

        domain = extract_unique_domain(o)

        if domain.ufl_cell().topological_dimension() < 3:
            raise ValueError("Facet edge lengths only make sense for topological dimension >= 3.")

        elif domain.ufl_coordinate_element().embedded_subdegree > 1:
            # Don't lower bendy cells, instead leave it to form compiler
            warnings.warn("Only know how to compute facet edge lengths of P1 or Q1 cell.")
            return o

        else:
            # P1 tetrahedron or Q1 hexahedron
            edges = FacetEdgeVectors(domain)
            num_edges = edges.ufl_shape[0]
            j = Index()
            elen2 = [real(edges[e, j] * conj(edges[e, j])) for e in range(num_edges)]
            return real(sqrt(reduce(reduction_op, elen2)))

    @memoized_handler
    def cell_normal(self, o):
        """Apply to cell_normal."""
        if self._preserve_types[o._ufl_typecode_]:
            return o

        domain = extract_unique_domain(o)
        gdim = domain.geometric_dimension()
        tdim = domain.topological_dimension()

        if tdim == gdim - 1:  # n-manifold embedded in n-1 space
            i = Index()
            J = self.jacobian(Jacobian(domain))

            if tdim == 2:
                # Surface in 3D
                t0 = as_vector(J[i, 0], i)
                t1 = as_vector(J[i, 1], i)
                cell_normal = cross_expr(t0, t1)
            elif tdim == 1:
                # Line in 2D (cell normal is 'up' for a line pointing
                # to the 'right')
                cell_normal = as_vector((-J[1, 0], J[0, 0]))
            else:
                raise ValueError(f"Cell normal not implemented for tdim {tdim}, gdim {gdim}")

            # Return normalized vector, sign corrected by cell
            # orientation
            co = CellOrientation(domain)
            return co * cell_normal / sqrt(cell_normal[i] * cell_normal[i])
        else:
            raise ValueError(f"Cell normal undefined for tdim {tdim}, gdim {gdim}")

    @memoized_handler
    def facet_normal(self, o):
        """Apply to facet_normal."""
        if self._preserve_types[o._ufl_typecode_]:
            return o

        domain = extract_unique_domain(o)
        tdim = domain.topological_dimension()

        if tdim == 1:
            # Special-case 1D (possibly immersed), for which we say
            # that n is just in the direction of J.
            J = self.jacobian(Jacobian(domain))  # dx/dX
            ndir = J[:, 0]

            gdim = domain.geometric_dimension()
            if gdim == 1:
                nlen = abs(ndir[0])
            else:
                i = Index()
                nlen = sqrt(ndir[i] * ndir[i])

            rn = ReferenceNormal(domain)  # +/- 1.0 here
            n = rn[0] * ndir / nlen
            r = n
        else:
            # Recall that the covariant Piola transform u -> J^(-T)*u
            # preserves tangential components. The normal vector is
            # characterised by having zero tangential component in
            # reference and physical space.
            Jinv = self.jacobian_inverse(JacobianInverse(domain))
            i, j = indices(2)

            rn = ReferenceNormal(domain)
            # compute signed, unnormalised normal; note transpose
            ndir = as_vector(Jinv[j, i] * rn[j], i)

            # normalise
            i = Index()
            n = ndir / sqrt(ndir[i] * ndir[i])
            r = n

        if r.ufl_shape != o.ufl_shape:
            raise ValueError(
                f"Inconsistent dimensions (in={o.ufl_shape[0]}, out={r.ufl_shape[0]})."
            )
        return r


def apply_geometry_lowering(form, preserve_types=()):
    """Change GeometricQuantity objects in expression to the lowest level GeometricQuantity objects.

    Assumes the expression is preprocessed or at least that derivatives have been expanded.

    Args:
        form: An Expr or Form.
        preserve_types: Preserved types
    """
    if isinstance(form, Form):
        newintegrals = [
            apply_geometry_lowering(integral, preserve_types) for integral in form.integrals()
        ]
        return Form(newintegrals)

    elif isinstance(form, Integral):
        integral = form
        if integral.integral_type() in (custom_integral_types + point_integral_types):
            automatic_preserve_types = [SpatialCoordinate, Jacobian]
        else:
            automatic_preserve_types = [CellCoordinate]
        preserve_types = set(preserve_types) | set(automatic_preserve_types)

        mf = GeometryLoweringApplier(preserve_types)
        newintegrand = map_expr_dag(mf, integral.integrand())
        return integral.reconstruct(integrand=newintegrand)

    elif isinstance(form, Expr):
        expr = form
        mf = GeometryLoweringApplier(preserve_types)
        return map_expr_dag(mf, expr)

    else:
        raise ValueError(f"Invalid type {form.__class__.__name__}")



"""Base form operator.

This module defines the BaseFormOperator class, which is the base class
for objects that can be seen as forms and as operators such as
ExternalOperator or Interpolate.
"""

# Copyright (C) 2019 Nacime Bouziani
#
# This file is part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
#
# Modified by Nacime Bouziani, 2021-2022

from collections import OrderedDict

from ufl.argument import Argument, Coargument
from ufl.constantvalue import as_ufl
from ufl.core.operator import Operator
from ufl.core.ufl_type import ufl_type
from ufl.form import BaseForm
from ufl.functionspace import AbstractFunctionSpace
from ufl.utils.counted import Counted


@ufl_type(num_ops="varying", is_differential=True)
class BaseFormOperator(Operator, BaseForm, Counted):
    """Base form operator."""

    # Slots are disabled here because they cause trouble in PyDOLFIN
    # multiple inheritance pattern:
    _ufl_noslots_ = True

    def __init__(self, *operands, function_space, derivatives=None, argument_slots=()):
        """Initialise.

        Args:
            operands: operands on which acts the operator.
            function_space: the FunctionSpace or MixedFunctionSpace on
                which to build this Function.
            derivatives: tuple specifying the derivative multiindex.
            argument_slots: tuple composed containing expressions with
                ufl.Argument or ufl.Coefficient objects.
        """
        BaseForm.__init__(self)
        ufl_operands = tuple(map(as_ufl, operands))
        argument_slots = tuple(map(as_ufl, argument_slots))
        Operator.__init__(self, ufl_operands)
        Counted.__init__(self, counted_class=BaseFormOperator)

        # -- Function space -- #
        if not isinstance(function_space, AbstractFunctionSpace):
            raise ValueError("Expecting a FunctionSpace or FiniteElement.")

        # -- Derivatives -- #
        # Some BaseFormOperator does have derivatives (e.g. ExternalOperator)
        # while other don't since they are fully determined by their
        # argument slots (e.g. Interpolate)
        self.derivatives = derivatives

        # -- Argument slots -- #
        if len(argument_slots) == 0:
            # Make v*
            v_star = Argument(function_space.dual(), 0)
            argument_slots = (v_star,)
        self._argument_slots = argument_slots

        # Internal variables for caching coefficient data
        self._coefficients = None

    # BaseFormOperators don't have free indices.
    ufl_free_indices = ()
    ufl_index_dimensions = ()

    def argument_slots(self, outer_form=False):
        """Returns a tuple of expressions containing argument and coefficient based expressions.

        We get an argument uhat when we take the Gateaux derivative in
        the direction uhat: d/du N(u; v*) = dNdu(u; uhat, v*) where uhat
        is a ufl.Argument and v* a ufl.Coargument Applying the action
        replace the last argument by coefficient: action(dNdu(u; uhat,
        v*), w) = dNdu(u; w, v*) where du is a ufl.Coefficient.
        """
        from ufl.algorithms.analysis import extract_arguments

        if not outer_form:
            return self._argument_slots
        # Takes into account argument contraction when a base form operator is in an outer form:
        # For example:
        #   F = N(u; v*) * v * dx can be seen as Action(v1 * v * dx, N(u; v*))
        #   => F.arguments() should return (v,)!
        return tuple(a for a in self._argument_slots[1:] if len(extract_arguments(a)) != 0)

    def coefficients(self):
        """Return all BaseCoefficient objects found in base form operator."""
        if self._coefficients is None:
            self._analyze_form_arguments()
        return self._coefficients

    def _analyze_form_arguments(self):
        """Analyze which Argument and Coefficient objects can be found in the base form."""
        from ufl.algorithms.analysis import extract_arguments, extract_coefficients, extract_type

        dual_arg, *arguments = self.argument_slots()
        # When coarguments are treated as BaseForms, they have two
        # arguments (one primal and one dual) as they map from V* to V*
        # => V* x V -> R. However, when they are treated as mere
        # "arguments", the primal space argument is discarded and we
        # only have the dual space argument (Coargument). This is the
        # exact same situation than BaseFormOperator's arguments which
        # are different depending on whether the BaseFormOperator is
        # used in an outer form or not.
        arguments = tuple(extract_type(dual_arg, Coargument)) + tuple(
            a for arg in arguments for a in extract_arguments(arg)
        )
        coefficients = tuple(c for op in self.ufl_operands for c in extract_coefficients(op))
        # Define canonical numbering of arguments and coefficients
        # 1) Need concept of order since we may have arguments with the same number
        #    because of form composition (`argument_slots(outer_form=True)`):
        #    Example: Let u \in V1 and N \in V2 and F = N(u; v*) * dx, then
        #    `derivative(F, u)` will contain dNdu(u; uhat, v*) with v* = Argument(0, V2)
        #    and uhat = Argument(0, V1) (since F.arguments() = ())
        # 2) Having sorted arguments also makes BaseFormOperator compatible with other
        #    BaseForm objects for which the highest-numbered argument always comes last.
        self._arguments = tuple(sorted(OrderedDict.fromkeys(arguments), key=lambda x: x.number()))
        self._coefficients = tuple(sorted(set(coefficients), key=lambda x: x.count()))

    def count(self):
        """Return the count associated to the base form operator."""
        return self._count

    @property
    def ufl_shape(self):
        """Return the UFL shape of the coefficient.produced by the operator."""
        arg, *_ = self.argument_slots()
        if isinstance(arg, BaseForm):
            arg, *_ = arg.arguments()
        return arg._ufl_shape

    def ufl_function_space(self):
        """Return the function space associated to the operator.

        I.e. return the dual of the base form operator's Coargument.
        """
        arg, *_ = self.argument_slots()
        if isinstance(arg, BaseForm):
            arg, *_ = arg.arguments()
            return arg._ufl_function_space
        return arg._ufl_function_space.dual()

    def _ufl_expr_reconstruct_(
        self, *operands, function_space=None, derivatives=None, argument_slots=None
    ):
        """Return a new object of the same type with new operands."""
        return type(self)(
            *operands,
            function_space=function_space or self.ufl_function_space(),
            derivatives=derivatives or self.derivatives,
            argument_slots=argument_slots or self.argument_slots(),
        )

    def __repr__(self):
        """Default repr string construction for base form operators."""
        r = f"{type(self).__name__}("
        r += ", ".join(repr(op) for op in self.ufl_operands)
        r += "; {self.ufl_function_space()!r}; "
        r += ", ".join(repr(arg) for arg in self.argument_slots())
        r += f"; derivatives={self.derivatives!r})"
        return r

    def __hash__(self):
        """Hash code for use in dicts."""
        hashdata = (
            type(self),
            tuple(hash(op) for op in self.ufl_operands),
            tuple(hash(arg) for arg in self._argument_slots),
            self.derivatives,
            hash(self.ufl_function_space()),
        )
        return hash(hashdata)

    def __eq__(self, other):
        """Check for equality."""
        raise NotImplementedError()



"""Operator."""
# Copyright (C) 2008-2016 Martin Sandve Alnæs
#
# This file is part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
#
# Modified by Anders Logg, 2008
# Modified by Massimiliano Leoni, 2016

from ufl.core.expr import Expr
from ufl.core.ufl_type import ufl_type


@ufl_type(is_abstract=True, is_terminal=False)
class Operator(Expr):
    """Base class for all operators, i.e. non-terminal expression types."""

    __slots__ = ("ufl_operands",)

    def __init__(self, operands=None):
        """Initialise."""
        Expr.__init__(self)

        # If operands is None, the type sets this itself. This is to
        # get around some tricky too-fancy __new__/__init__ design in
        # algebra.py, for now.  It would be nicer to make the classes
        # in algebra.py pass operands here.
        if operands is not None:
            self.ufl_operands = operands

    def _ufl_expr_reconstruct_(self, *operands):
        """Return a new object of the same type with new operands."""
        return self._ufl_class_(*operands)

    def _ufl_signature_data_(self):
        """Get UFL signature data."""
        return self._ufl_typecode_

    def _ufl_compute_hash_(self):
        """Compute a hash code for this expression. Used by sets and dicts."""
        return hash((self._ufl_typecode_,) + tuple(hash(o) for o in self.ufl_operands))

    def __repr__(self):
        """Default repr string construction for operators."""
        # This should work for most cases
        return f"{self._ufl_class_.__name__}({', '.join(repr(op) for op in self.ufl_operands)})"



"""A collection of utility algorithms for handling UFL files."""

# Copyright (C) 2008-2016 Martin Sandve Alnæs
#
# This file is part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
#
# Modified by Anders Logg, 2008-2009.
# Modified by Marie E. Rognes, 2011.

import io
import os
import re

from ufl.argument import Argument
from ufl.coefficient import Coefficient
from ufl.constant import Constant
from ufl.core.expr import Expr
from ufl.finiteelement import AbstractFiniteElement
from ufl.form import Form
from ufl.utils.sorting import sorted_by_key


class FileData(object):
    """File data."""

    def __init__(self):
        """Initialise."""
        self.elements = []
        self.coefficients = []
        self.expressions = []
        self.forms = []
        self.object_names = {}
        self.object_by_name = {}
        self.reserved_objects = {}

    def __bool__(self):
        """Convert to a bool."""
        return bool(
            self.elements
            or self.coefficients
            or self.forms
            or self.expressions
            or self.object_names
            or self.object_by_name
            or self.reserved_objects
        )

    __nonzero__ = __bool__


def read_lines_decoded(fn):
    """Read decoded lines of a UFL file."""
    r = re.compile(b".*coding: *([^ ]+)")

    def match(line):
        """Match."""
        return r.match(line, re.ASCII)

    # First read lines as bytes
    with io.open(fn, "rb") as f:
        lines = f.readlines()

    # Check for coding: in the first two lines
    for i in range(min(2, len(lines))):
        m = match(lines[i])
        if m:
            (encoding,) = m.groups()
            # Drop encoding line
            lines = lines[:i] + lines[i + 1 :]
            break
    else:
        # Default to utf-8 (works for ascii files
        # as well, default for python files in py3)
        encoding = "utf-8"

    # Decode all lines
    lines = [line.decode(encoding=encoding) for line in lines]
    return lines


def read_ufl_file(filename):
    """Read a UFL file."""
    if not os.path.exists(filename):
        raise ValueError(f"File '{filename}' doesn't exist.")
    lines = read_lines_decoded(filename)
    code = "".join(lines)
    return code


def execute_ufl_code(uflcode):
    """Execute code."""
    namespace = {}
    exec(uflcode, namespace)
    return namespace


def interpret_ufl_namespace(namespace):
    """Take a namespace dict from an executed ufl file and convert it to a FileData object."""
    # Object to hold all returned data
    ufd = FileData()

    # Extract object names for Form, Coefficient and AbstractFiniteElement objects
    # The use of id(obj) as key in object_names is necessary
    # because we need to distinguish between instances,
    # and not just between objects with different values.
    for name, value in sorted_by_key(namespace):
        # Store objects by reserved name OR instance id
        reserved_names = ("unknown",)  # Currently only one reserved name
        if name in reserved_names:
            # Store objects with reserved names
            ufd.reserved_objects[name] = value
            # FIXME: Remove after FFC is updated to use reserved_objects:
            ufd.object_names[name] = value
            ufd.object_by_name[name] = value
        elif isinstance(
            value, (AbstractFiniteElement, Coefficient, Constant, Argument, Form, Expr)
        ):
            # Store instance <-> name mappings for important objects
            # without a reserved name
            ufd.object_names[id(value)] = name
            ufd.object_by_name[name] = value

    # Get list of exported forms
    forms = namespace.get("forms")
    if forms is None:
        # Get forms from object_by_name, which has already mapped
        # tuple->Form where needed
        def get_form(name):
            form = ufd.object_by_name.get(name)
            return form if isinstance(form, Form) else None

        a_form = get_form("a")
        L_form = get_form("L")
        M_form = get_form("M")
        forms = [a_form, L_form, M_form]
        # Add forms F and J if not "a" and "L" are used
        if a_form is None or L_form is None:
            F_form = get_form("F")
            J_form = get_form("J")
            forms += [F_form, J_form]
        # Remove Nones
        forms = [f for f in forms if isinstance(f, Form)]
    ufd.forms = forms

    # Validate types
    if not isinstance(ufd.forms, (list, tuple)):
        raise ValueError(f"Expecting 'forms' to be a list or tuple, not '{type(ufd.forms)}'.")
    if not all(isinstance(a, Form) for a in ufd.forms):
        raise ValueError("Expecting 'forms' to be a list of Form instances.")

    # Get list of exported elements
    elements = namespace.get("elements")
    if elements is None:
        elements = [ufd.object_by_name.get(name) for name in ("element",)]
        elements = [e for e in elements if e is not None]
    ufd.elements = elements

    # Validate types
    if not isinstance(ufd.elements, (list, tuple)):
        raise ValueError(
            f"Expecting 'elements' to be a list or tuple, not '{type(ufd.elements)}''."
        )
    if not all(isinstance(e, AbstractFiniteElement) for e in ufd.elements):
        raise ValueError("Expecting 'elements' to be a list of AbstractFiniteElement instances.")

    # Get list of exported coefficients
    functions = []
    ufd.coefficients = namespace.get("coefficients", functions)

    # Validate types
    if not isinstance(ufd.coefficients, (list, tuple)):
        raise ValueError(
            f"Expecting 'coefficients' to be a list or tuple, not '{type(ufd.coefficients)}'."
        )
    if not all(isinstance(e, Coefficient) for e in ufd.coefficients):
        raise ValueError("Expecting 'coefficients' to be a list of Coefficient instances.")

    # Get list of exported expressions
    ufd.expressions = namespace.get("expressions", [])

    # Validate types
    if not isinstance(ufd.expressions, (list, tuple)):
        raise ValueError(
            f"Expecting 'expressions' to be a list or tuple, not '{type(ufd.expressions)}'."
        )
    if not all(isinstance(e[0], Expr) for e in ufd.expressions):
        raise ValueError("Expecting 'expressions' to be a list of Expr instances.")

    # Return file data
    return ufd


def load_ufl_file(filename):
    """Load a UFL file with elements, coefficients, expressions and forms."""
    # Read code from file and execute it
    uflcode = read_ufl_file(filename)
    namespace = execute_ufl_code(uflcode)
    return interpret_ufl_namespace(namespace)


def load_forms(filename):
    """Return a list of all forms in a file."""
    ufd = load_ufl_file(filename)
    return ufd.forms



"""Protocols."""
# Copyright (C) 2008-2016 Martin Sandve Alnæs
#
# This file is part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later


def id_or_none(obj):
    """Returns None if the object is None, obj.ufl_id() if available, or id(obj) if not.

    This allows external libraries to implement an alternative
    to id(obj) in the ufl_id() function, such that ufl can identify
    objects as the same without knowing about their types.
    """
    if obj is None:
        return None
    elif hasattr(obj, "ufl_id"):
        return obj.ufl_id()
    else:
        return id(obj)



"""This module defines the UFL finite element classes."""
# Copyright (C) 2008-2016 Martin Sandve Alnæs
#
# This file is part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
#
# Modified by Kristian B. Oelgaard
# Modified by Marie E. Rognes 2010, 2012
# Modified by Massimiliano Leoni, 2016
# Modified by Matthew Scroggs, 2023

from __future__ import annotations

import abc as _abc
import typing as _typing

from ufl.cell import Cell as _Cell
from ufl.pullback import AbstractPullback as _AbstractPullback
from ufl.pullback import IdentityPullback as _IdentityPullback
from ufl.pullback import MixedPullback as _MixedPullback
from ufl.pullback import SymmetricPullback as _SymmetricPullback
from ufl.sobolevspace import SobolevSpace as _SobolevSpace
from ufl.utils.sequences import product

__all_classes__ = ["AbstractFiniteElement", "FiniteElement", "MixedElement", "SymmetricElement"]


class AbstractFiniteElement(_abc.ABC):
    """Base class for all finite elements.

    To make your element library compatible with UFL, you should make a
    subclass of AbstractFiniteElement and provide implementions of all
    the abstract methods and properties. All methods and properties that
    are not marked as abstract are implemented here and should not need
    to be overwritten in your subclass.

    An example of how the methods in your subclass could be implemented
    can be found in Basix; see
    https://github.com/FEniCS/basix/blob/main/python/basix/ufl.py
    """

    @_abc.abstractmethod
    def __repr__(self) -> str:
        """Format as string for evaluation as Python object."""

    @_abc.abstractmethod
    def __str__(self) -> str:
        """Format as string for nice printing."""

    @_abc.abstractmethod
    def __hash__(self) -> int:
        """Return a hash."""

    @_abc.abstractmethod
    def __eq__(self, other: AbstractFiniteElement) -> bool:
        """Check if this element is equal to another element."""

    @_abc.abstractproperty
    def sobolev_space(self) -> _SobolevSpace:
        """Return the underlying Sobolev space."""

    @_abc.abstractproperty
    def pullback(self) -> _AbstractPullback:
        """Return the pullback for this element."""

    @_abc.abstractproperty
    def embedded_superdegree(self) -> _typing.Union[int, None]:
        """Degree of the minimum degree Lagrange space that spans this element.

        This returns the degree of the lowest degree Lagrange space such
        that the polynomial space of the Lagrange space is a superspace
        of this element's polynomial space. If this element contains
        basis functions that are not in any Lagrange space, this
        function should return None.

        Note that on a simplex cells, the polynomial space of Lagrange
        space is a complete polynomial space, but on other cells this is
        not true. For example, on quadrilateral cells, the degree 1
        Lagrange space includes the degree 2 polynomial xy.
        """

    @_abc.abstractproperty
    def embedded_subdegree(self) -> int:
        """Degree of the maximum degree Lagrange space that is spanned by this element.

        This returns the degree of the highest degree Lagrange space
        such that the polynomial space of the Lagrange space is a
        subspace of this element's polynomial space. If this element's
        polynomial space does not include the constant function, this
        function should return -1.

        Note that on a simplex cells, the polynomial space of Lagrange
        space is a complete polynomial space, but on other cells this is
        not true. For example, on quadrilateral cells, the degree 1
        Lagrange space includes the degree 2 polynomial xy.
        """

    @_abc.abstractproperty
    def cell(self) -> _Cell:
        """Return the cell of the finite element."""

    @_abc.abstractproperty
    def reference_value_shape(self) -> _typing.Tuple[int, ...]:
        """Return the shape of the value space on the reference cell."""

    @_abc.abstractproperty
    def sub_elements(self) -> _typing.List:
        """Return list of sub-elements.

        This function does not recurse: ie it does not extract the sub-elements
        of sub-elements.
        """

    def __ne__(self, other: AbstractFiniteElement) -> bool:
        """Check if this element is different to another element."""
        return not self.__eq__(other)

    def is_cellwise_constant(self) -> bool:
        """Check whether this element is spatially constant over each cell."""
        return self.embedded_superdegree == 0

    def _ufl_hash_data_(self) -> str:
        """Return UFL hash data."""
        return repr(self)

    def _ufl_signature_data_(self) -> str:
        """Return UFL signature data."""
        return repr(self)

    @property
    def reference_value_size(self) -> int:
        """Return the integer product of the reference value shape."""
        return product(self.reference_value_shape)

    @property
    def num_sub_elements(self) -> int:
        """Return number of sub-elements.

        This function does not recurse: ie it does not count the sub-elements of
        sub-elements.
        """
        return len(self.sub_elements)


class FiniteElement(AbstractFiniteElement):
    """A directly defined finite element."""

    __slots__ = (
        "_repr",
        "_str",
        "_family",
        "_cell",
        "_degree",
        "_reference_value_shape",
        "_pullback",
        "_sobolev_space",
        "_sub_elements",
        "_subdegree",
    )

    def __init__(
        self,
        family: str,
        cell: _Cell,
        degree: int,
        reference_value_shape: _typing.Tuple[int, ...],
        pullback: _AbstractPullback,
        sobolev_space: _SobolevSpace,
        sub_elements=[],
        _repr: _typing.Optional[str] = None,
        _str: _typing.Optional[str] = None,
        subdegree: _typing.Optional[int] = None,
    ):
        """Initialise a finite element.

        This class should only be used for testing

        Args:
            family: The family name of the element
            cell: The cell on which the element is defined
            degree: The polynomial degree of the element
            reference_value_shape: The reference value shape of the element
            pullback: The pullback to use
            sobolev_space: The Sobolev space containing this element
            sub_elements: Sub elements of this element
            _repr: A string representation of this elements
            _str: A string for printing
            subdegree: The embedded subdegree of this element
        """
        if subdegree is None:
            self._subdegree = degree
        else:
            self._subdegree = subdegree
        if _repr is None:
            if len(sub_elements) > 0:
                self._repr = (
                    f'ufl.finiteelement.FiniteElement("{family}", {cell}, {degree}, '
                    f"{reference_value_shape}, {pullback}, {sobolev_space}, {sub_elements!r})"
                )
            else:
                self._repr = (
                    f'ufl.finiteelement.FiniteElement("{family}", {cell}, {degree}, '
                    f"{reference_value_shape}, {pullback}, {sobolev_space})"
                )
        else:
            self._repr = _repr
        if _str is None:
            self._str = f"<{family}{degree} on a {cell}>"
        else:
            self._str = _str
        self._family = family
        self._cell = cell
        self._degree = degree
        self._reference_value_shape = reference_value_shape
        self._pullback = pullback
        self._sobolev_space = sobolev_space
        self._sub_elements = sub_elements

    def __repr__(self) -> str:
        """Format as string for evaluation as Python object."""
        return self._repr

    def __str__(self) -> str:
        """Format as string for nice printing."""
        return self._str

    def __hash__(self) -> int:
        """Return a hash."""
        return hash(f"{self!r}")

    def __eq__(self, other) -> bool:
        """Check if this element is equal to another element."""
        return type(self) is type(other) and repr(self) == repr(other)

    @property
    def sobolev_space(self) -> _SobolevSpace:
        """Return the underlying Sobolev space."""
        return self._sobolev_space

    @property
    def pullback(self) -> _AbstractPullback:
        """Return the pullback for this element."""
        return self._pullback

    @property
    def embedded_superdegree(self) -> _typing.Union[int, None]:
        """Degree of the minimum degree Lagrange space that spans this element.

        This returns the degree of the lowest degree Lagrange space such
        that the polynomial space of the Lagrange space is a superspace
        of this element's polynomial space. If this element contains
        basis functions that are not in any Lagrange space, this
        function should return None.

        Note that on a simplex cells, the polynomial space of Lagrange
        space is a complete polynomial space, but on other cells this is
        not true. For example, on quadrilateral cells, the degree 1
        Lagrange space includes the degree 2 polynomial xy.
        """
        return self._degree

    @property
    def embedded_subdegree(self) -> int:
        """Degree of the maximum degree Lagrange space that is spanned by this element.

        This returns the degree of the highest degree Lagrange space
        such that the polynomial space of the Lagrange space is a
        subspace of this element's polynomial space. If this element's
        polynomial space does not include the constant function, this
        function should return -1.

        Note that on a simplex cells, the polynomial space of Lagrange
        space is a complete polynomial space, but on other cells this is
        not true. For example, on quadrilateral cells, the degree 1
        Lagrange space includes the degree 2 polynomial xy.
        """
        return self._subdegree

    @property
    def cell(self) -> _Cell:
        """Return the cell of the finite element."""
        return self._cell

    @property
    def reference_value_shape(self) -> _typing.Tuple[int, ...]:
        """Return the shape of the value space on the reference cell."""
        return self._reference_value_shape

    @property
    def sub_elements(self) -> _typing.List:
        """Return list of sub-elements.

        This function does not recurse: ie it does not extract the
        sub-elements of sub-elements.
        """
        return self._sub_elements


class SymmetricElement(FiniteElement):
    """A symmetric finite element."""

    def __init__(
        self,
        symmetry: _typing.Dict[_typing.Tuple[int, ...], int],
        sub_elements: _typing.List[AbstractFiniteElement],
    ):
        """Initialise a symmetric element.

        This class should only be used for testing

        Args:
            symmetry: Map from physical components to reference components
            sub_elements: Sub-elements of this element
        """
        self._sub_elements = sub_elements
        pullback = _SymmetricPullback(self, symmetry)
        reference_value_shape = (sum(e.reference_value_size for e in sub_elements),)
        degree = max(e.embedded_superdegree for e in sub_elements)
        cell = sub_elements[0].cell
        for e in sub_elements:
            if e.cell != cell:
                raise ValueError("All sub-elements must be defined on the same cell")
        sobolev_space = max(e.sobolev_space for e in sub_elements)

        super().__init__(
            "Symmetric element",
            cell,
            degree,
            reference_value_shape,
            pullback,
            sobolev_space,
            sub_elements=sub_elements,
            _repr=(f"ufl.finiteelement.SymmetricElement({symmetry!r}, {sub_elements!r})"),
            _str=f"<symmetric element on a {cell}>",
        )


class MixedElement(FiniteElement):
    """A mixed element."""

    def __init__(self, sub_elements):
        """Initialise a mixed element.

        This class should only be used for testing

        Args:
            sub_elements: Sub-elements of this element
        """
        sub_elements = [MixedElement(e) if isinstance(e, list) else e for e in sub_elements]
        cell = sub_elements[0].cell
        for e in sub_elements:
            assert e.cell == cell
        degree = max(e.embedded_superdegree for e in sub_elements)
        reference_value_shape = (sum(e.reference_value_size for e in sub_elements),)
        if all(isinstance(e.pullback, _IdentityPullback) for e in sub_elements):
            pullback = _IdentityPullback()
        else:
            pullback = _MixedPullback(self)
        sobolev_space = max(e.sobolev_space for e in sub_elements)

        super().__init__(
            "Mixed element",
            cell,
            degree,
            reference_value_shape,
            pullback,
            sobolev_space,
            sub_elements=sub_elements,
            _repr=f"ufl.finiteelement.MixedElement({sub_elements!r})",
            _str=f"<MixedElement with {len(sub_elements)} sub-element(s)>",
        )



# Copyright (C) 2006-2007 Kristiand Oelgaard and Anders Logg
#
# This file is part of UFL.
#
# UFL is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# UFL is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with UFL. If not, see <http://www.gnu.org/licenses/>.
#
# First added:  2006-12-05
# Last changed: 2007-07-15
#
# The bilinear form a(v, u) and linear form L(v) for
# Poisson's equation in a discontinuous Galerkin (DG)
# formulation.
from ufl import (
    Coefficient,
    Constant,
    FacetNormal,
    FunctionSpace,
    Mesh,
    TestFunction,
    TrialFunction,
    avg,
    dot,
    dS,
    ds,
    dx,
    grad,
    inner,
    jump,
    triangle,
)
from ufl.finiteelement import FiniteElement
from ufl.pullback import identity_pullback
from ufl.sobolevspace import H1, L2

cell = triangle
element = FiniteElement("Discontinuous Lagrange", cell, 1, (), identity_pullback, L2)
domain = Mesh(FiniteElement("Lagrange", cell, 1, (2,), identity_pullback, H1))
space = FunctionSpace(domain, element)

v = TestFunction(space)
u = TrialFunction(space)
f = Coefficient(space)

n = FacetNormal(domain)
h = Constant(domain)

gN = Coefficient(space)

alpha = 4.0
gamma = 8.0

a = (
    inner(grad(v), grad(u)) * dx
    - inner(avg(grad(v)), jump(u, n)) * dS
    - inner(jump(v, n), avg(grad(u))) * dS
    + alpha / h("+") * dot(jump(v, n), jump(u, n)) * dS
    - inner(grad(v), u * n) * ds
    - inner(v * n, grad(u)) * ds
    + gamma / h * v * u * ds
)

L = v * f * dx + v * gN * ds



"""Utilities for analysing and manipulating free index tuples."""
# Copyright (C) 2008-2016 Martin Sandve Alnæs
#
# This file is part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later

from ufl.core.multiindex import FixedIndex, Index, indices

# FIXME: Some of these might be merged into one function, some might
# be optimized


def unique_sorted_indices(indices):
    """Get unique sorted indices.

    Given a list of (id, dim) tuples already sorted by id,
    return a unique list with duplicates removed.
    Also checks that the dimensions of duplicates are matching.
    """
    newindices = []
    prev = (None, None)
    for i in indices:
        if i[0] != prev[0]:
            newindices.append(i)
            prev = i
        else:
            if i[1] != prev[1]:
                raise ValueError("Nonmatching dimensions for free indices with same id!")
    return tuple(newindices)


def merge_unique_indices(afi, afid, bfi, bfid):
    """Merge two pairs of (index ids, index dimensions) sequences into one pair without duplicates.

    The id tuples afi, bfi are assumed already sorted by id.
    Given a list of (id, dim) tuples already sorted by id,
    return a unique list with duplicates removed.
    Also checks that the dimensions of duplicates are matching.
    """
    na = len(afi)
    nb = len(bfi)

    if na == 0:
        return bfi, bfid
    elif nb == 0:
        return afi, afid

    ak = 0
    bk = 0

    fi = []
    fid = []

    while True:
        if afi[ak] < bfi[bk]:
            fi.append(afi[ak])
            fid.append(afid[ak])
            ak += 1
        elif afi[ak] > bfi[bk]:
            fi.append(bfi[bk])
            fid.append(bfid[bk])
            bk += 1
        else:
            fi.append(afi[ak])
            fid.append(afid[ak])
            ak += 1
            bk += 1

        if ak == na:
            if bk != nb:
                fi.extend(bfi[bk:])
                fid.extend(bfid[bk:])
            break
        elif bk == nb:
            fi.extend(afi[ak:])
            fid.extend(afid[ak:])
            break

    return tuple(fi), tuple(fid)


def remove_indices(fi, fid, rfi):
    """Remove indices."""
    if not rfi:
        return fi, fid

    rfip = sorted((r, p) for p, r in enumerate(rfi))

    nrfi = len(rfi)
    nfi = len(fi)
    shape = [None] * nrfi
    k = 0
    pos = 0
    newfiid = []
    while pos < nfi:
        rk = rfip[k][0]

        # Keep
        while fi[pos] < rk:
            newfiid.append((fi[pos], fid[pos]))
            pos += 1

        # Skip
        removed = 0
        while pos < nfi and fi[pos] == rk:
            shape[rfip[k][1]] = fid[pos]
            pos += 1
            removed += 1

        # Expecting to find each index from rfi in fi
        if not removed:
            raise ValueError(f"Index to be removed ({rk}) not part of indices ({fi}).")

        # Next to remove
        k += 1
        if k == nrfi:
            # No more to remove, keep the rest
            if pos < nfi:
                newfiid.extend(zip(fi[pos:], fid[pos:]))
            break

    assert None not in shape

    # Unpack into two tuples
    fi, fid = zip(*newfiid) if newfiid else ((), ())

    return fi, fid, tuple(shape)


def create_slice_indices(component, shape, fi):
    """Create slice indices."""
    all_indices = []
    slice_indices = []
    repeated_indices = []
    free_indices = []

    for ind in component:
        if isinstance(ind, Index):
            all_indices.append(ind)
            if ind.count() in fi or ind in free_indices:
                repeated_indices.append(ind)
            free_indices.append(ind)
        elif isinstance(ind, FixedIndex):
            if int(ind) >= shape[len(all_indices)]:
                raise ValueError("Index out of bounds.")
            all_indices.append(ind)
        elif isinstance(ind, int):
            if int(ind) >= shape[len(all_indices)]:
                raise ValueError("Index out of bounds.")
            all_indices.append(FixedIndex(ind))
        elif isinstance(ind, slice):
            if ind != slice(None):
                raise ValueError("Only full slices (:) allowed.")
            i = Index()
            slice_indices.append(i)
            all_indices.append(i)
        elif ind == Ellipsis:
            er = len(shape) - len(component) + 1
            ii = indices(er)
            slice_indices.extend(ii)
            all_indices.extend(ii)
        else:
            raise ValueError(f"Not expecting {ind}.")

    if len(all_indices) != len(shape):
        raise ValueError("Component and shape length don't match.")

    return tuple(all_indices), tuple(slice_indices), tuple(repeated_indices)


# Outer etc.
def merge_nonoverlapping_indices(a, b):
    """Merge non-overlapping free indices into one representation.

    Example:
      C[i,j,r,s] = outer(A[i,s], B[j,r])
      A, B -> (i,j,r,s), (idim,jdim,rdim,sdim)
    """
    # Extract input properties
    ai = a.ufl_free_indices
    bi = b.ufl_free_indices
    aid = a.ufl_index_dimensions
    bid = b.ufl_index_dimensions

    # Merge lists to return
    s = sorted(zip(ai + bi, aid + bid))
    if s:
        free_indices, index_dimensions = zip(*s)
        # Consistency checks
        if len(set(free_indices)) != len(free_indices):
            raise ValueError("Not expecting repeated indices.")
    else:
        free_indices, index_dimensions = (), ()
    return free_indices, index_dimensions


# Product
def merge_overlapping_indices(afi, afid, bfi, bfid):
    """Merge overlapping free indices into one free and one repeated representation.

    Example:
      C[j,r] := A[i,j,k] * B[i,r,k]
      A, B -> (j,r), (jdim,rdim), (i,k), (idim,kdim)
    """
    # Extract input properties
    an = len(afi)
    bn = len(bfi)

    # Lists to return
    free_indices = []
    index_dimensions = []
    repeated_indices = []
    repeated_index_dimensions = []

    # Find repeated indices, brute force version
    for i0 in range(an):
        for i1 in range(bn):
            if afi[i0] == bfi[i1]:
                repeated_indices.append(afi[i0])
                repeated_index_dimensions.append(afid[i0])
                break

    # Collect only non-repeated indices, brute force version
    for i, d in sorted(zip(afi + bfi, afid + bfid)):
        if i not in repeated_indices:
            free_indices.append(i)
            index_dimensions.append(d)

    # Consistency checks
    if len(set(free_indices)) != len(free_indices):
        raise ValueError("Not expecting repeated indices left.")
    if len(free_indices) + 2 * len(repeated_indices) != an + bn:
        raise ValueError("Expecting only twice repeated indices.")

    return (
        tuple(free_indices),
        tuple(index_dimensions),
        tuple(repeated_indices),
        tuple(repeated_index_dimensions),
    )



"""Base class for multifunctions with UFL ``Expr`` type dispatch."""
# Copyright (C) 2008-2016 Martin Sandve Alnæs
#
# This file is part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
#
# Modified by Massimiliano Leoni, 2016

import inspect

from ufl.core.expr import Expr
from ufl.core.ufl_type import UFLType


def get_num_args(function):
    """Return the number of arguments accepted by *function*."""
    sig = inspect.signature(function)
    return len(sig.parameters) + 1


def memoized_handler(handler):
    """Function decorator to memoize ``MultiFunction`` handlers."""

    def _memoized_handler(self, o):
        c = getattr(self, "_memoized_handler_cache")
        r = c.get(o)
        if r is None:
            r = handler(self, o)
            c[o] = r
        return r

    return _memoized_handler


class MultiFunction(object):
    """Base class for collections of non-recursive expression node handlers.

    Subclass this (remember to call the ``__init__`` method of this class),
    and implement handler functions for each ``Expr`` type, using the lower case
    handler name of the type (``exprtype._ufl_handler_name_``).

    This class is optimized for efficient type based dispatch in the
    ``__call__``
    operator via typecode based lookup of the handler function bound to the
    algorithm object. Of course Python's function call overhead still applies.
    """

    _handlers_cache = {}

    def __init__(self):
        """Initialise."""
        # Analyse class properties and cache handler data the
        # first time this is run for a particular class
        # (cached for each algorithm for performance)
        algorithm_class = type(self)
        cache_data = MultiFunction._handlers_cache.get(algorithm_class)
        if not cache_data:
            handler_names = [None] * len(Expr._ufl_all_classes_)

            # Iterate over the inheritance chain for each Expr
            # subclass (NB! This assumes that all UFL classes inherits
            # from a single Expr subclass and that the first
            # superclass is always from the UFL Expr hierarchy!)
            for classobject in Expr._ufl_all_classes_:
                for c in classobject.mro():
                    # Register classobject with handler for the first
                    # encountered superclass
                    try:
                        handler_name = c._ufl_handler_name_
                    except AttributeError as attribute_error:
                        if type(classobject) is not UFLType:
                            raise attribute_error
                        # Default handler name for UFL types
                        handler_name = UFLType._ufl_handler_name_

                    if hasattr(self, handler_name):
                        handler_names[classobject._ufl_typecode_] = handler_name
                        break
            is_cutoff_type = [get_num_args(getattr(self, name)) == 2 for name in handler_names]
            cache_data = (handler_names, is_cutoff_type)
            MultiFunction._handlers_cache[algorithm_class] = cache_data

        # Build handler list for this particular class (get functions
        # bound to self, these cannot be cached)
        handler_names, is_cutoff_type = cache_data
        self._handlers = [getattr(self, name) for name in handler_names]
        self._is_cutoff_type = is_cutoff_type

        # Create cache for memoized_handler
        self._memoized_handler_cache = {}

    def __call__(self, o, *args):
        """Delegate to handler function based on typecode of first argument."""
        return self._handlers[o._ufl_typecode_](o, *args)

    def undefined(self, o, *args):
        """Trigger error for types with missing handlers."""
        raise ValueError(f"No handler defined for {o._ufl_class_.__name__}.")

    def reuse_if_untouched(self, o, *ops):
        """Reuse object if operands are the same objects.

        Use in your own subclass by setting e.g.
        ::

            expr = MultiFunction.reuse_if_untouched

        as a default rule.
        """
        if all(a is b for a, b in zip(o.ufl_operands, ops)):
            return o
        else:
            return o._ufl_expr_reconstruct_(*ops)

    # Set default behaviour for any UFLType as undefined
    ufl_type = undefined



"""Test tensor algebra operators."""

import pytest

from ufl import (
    FacetNormal,
    Mesh,
    as_matrix,
    as_tensor,
    as_vector,
    cofac,
    cross,
    det,
    dev,
    diag,
    diag_vector,
    dot,
    inner,
    inv,
    outer,
    perp,
    skew,
    sym,
    tr,
    transpose,
    triangle,
    zero,
)
from ufl.algorithms.remove_complex_nodes import remove_complex_nodes
from ufl.finiteelement import FiniteElement
from ufl.pullback import identity_pullback
from ufl.sobolevspace import H1


@pytest.fixture(scope="module")
def A():
    return as_matrix([[2, 3], [4, 5]])


@pytest.fixture(scope="module")
def B():
    return as_matrix([[6, 7], [8, 9]])


@pytest.fixture(scope="module")
def u():
    return as_vector([10, 20])


@pytest.fixture(scope="module")
def v():
    return as_vector([30, 40])


def test_repeated_as_tensor(self, A, B, u, v):
    A2 = as_tensor(A)
    B2 = as_matrix(B)
    u2 = as_tensor(u)
    v2 = as_vector(v)
    assert A2 == A
    assert B2 == B
    assert u2 == u
    assert v2 == v


def test_outer(self, A, B, u, v):
    C = outer(u, v)
    D = as_matrix([[10 * 30, 10 * 40], [20 * 30, 20 * 40]])
    self.assertEqualValues(C, D)

    C = outer(A, v)
    A, v = A, v
    dims = (0, 1)
    D = as_tensor([[[A[i, j] * v[k] for k in dims] for j in dims] for i in dims])
    self.assertEqualValues(C, D)

    # TODO: Test other ranks


def test_inner(self, A, B, u, v):
    C = inner(A, B)
    D = 2 * 6 + 3 * 7 + 4 * 8 + 5 * 9
    self.assertEqualValues(C, D)

    C = inner(u, v)
    D = 10 * 30 + 20 * 40
    self.assertEqualValues(C, D)


def test_pow2_inner(self, A, u):
    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    f = FacetNormal(domain)[0]
    f2 = f * f
    assert f2 == remove_complex_nodes(inner(f, f))

    u2 = u**2
    assert u2 == remove_complex_nodes(inner(u, u))

    A2 = A**2
    assert A2 == remove_complex_nodes(inner(A, A))

    # Only tensor**2 notation is supported:
    self.assertRaises(BaseException, lambda: A**3)


def test_dot(self, A, B, u, v):
    C = dot(u, v)
    D = 10 * 30 + 20 * 40
    self.assertEqualValues(C, D)

    C = dot(A, B)
    dims = (0, 1)
    D = as_matrix([[sum(A[i, k] * B[k, j] for k in dims) for j in dims] for i in dims])
    self.assertEqualValues(C, D)


def test_cross(self):
    u = as_vector([3, 3, 3])
    v = as_vector([2, 2, 2])
    C = cross(u, v)
    D = zero(3)
    self.assertEqualValues(C, D)

    u = as_vector([3, 3, 0])
    v = as_vector([-2, 2, 0])
    C = cross(u, v)
    z = det(as_matrix([[3, 3], [-2, 2]]))
    D = as_vector([0, 0, z])
    self.assertEqualValues(C, D)


def test_perp(self):
    # Test perp is generally doing the correct thing
    u = as_vector([3, 1])
    v = perp(u)
    w = as_vector([-1, 3])
    self.assertEqualValues(v, w)

    # Test that a perp does the correct thing to Zero
    u = zero(2)
    v = perp(u)
    self.assertEqualValues(u, v)

    # Test that perp throws an error if the wrong thing is provided
    u = as_vector([3, 1, -1])  # 3D vector instead of 2D
    with pytest.raises(ValueError):
        v = perp(u)
    u = as_matrix([[1, 3], [0, 4]])  # Matrix instead of vector
    with pytest.raises(ValueError):
        v = perp(u)


def xtest_dev(self, A):
    C = dev(A)
    D = 0 * C  # FIXME: Add expected value here
    self.assertEqualValues(C, D)


def test_skew(self, A):
    C = skew(A)
    A, dims = A, (0, 1)
    D = 0.5 * as_matrix([[A[i, j] - A[j, i] for j in dims] for i in dims])
    self.assertEqualValues(C, D)


def test_sym(self, A):
    C = sym(A)
    A, dims = A, (0, 1)
    D = 0.5 * as_matrix([[A[i, j] + A[j, i] for j in dims] for i in dims])
    self.assertEqualValues(C, D)


def test_transpose(self, A):
    C = transpose(A)
    dims = (0, 1)
    D = as_matrix([[A[j, i] for j in dims] for i in dims])
    self.assertEqualValues(C, D)


def test_diag(self, A, u):
    dims = (0, 1)

    C = diag(A)
    D = as_matrix([[(0 if i != j else A[i, i]) for j in dims] for i in dims])
    self.assertEqualValues(C, D)

    C = diag(u)
    D = as_matrix([[(0 if i != j else u[i]) for j in dims] for i in dims])
    self.assertEqualValues(C, D)


def test_diag_vector(self, A):
    dims = (0, 1)
    C = diag_vector(A)
    D = as_vector([A[i, i] for i in dims])
    self.assertEqualValues(C, D)


def test_tr(self, A):
    C = tr(A)
    A, dims = A, (0, 1)
    D = sum(A[i, i] for i in dims)
    self.assertEqualValues(C, D)


def test_det(self, A):
    dims = (0, 1)
    C = det(A)
    D = sum((-A[i, 0] * A[0, i] if i != 0 else A[i - 1, -1] * A[i, 0]) for i in dims)
    self.assertEqualValues(C, D)


def test_cofac(self, A):
    C = cofac(A)
    D = as_matrix([[(-A[i, j] if i != j else A[i, j]) for j in (-1, 0)] for i in (-1, 0)])
    self.assertEqualValues(C, D)


def xtest_inv(self, A):
    # FIXME: Test fails probably due to integer division
    C = inv(A)
    detA = sum((-A[i, 0] * A[0, i] if i != 0 else A[i - 1, -1] * A[i, 0]) for i in (0, 1))
    D = as_matrix([[(-A[i, j] if i != j else A[i, j]) for j in (-1, 0)] for i in (-1, 0)]) / detA
    self.assertEqualValues(C, D)



"""Types for representing a geometric domain."""

# Copyright (C) 2008-2016 Martin Sandve Alnæs
#
# This file is part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later

import numbers

from ufl.cell import AbstractCell
from ufl.core.ufl_id import attach_ufl_id
from ufl.core.ufl_type import UFLObject
from ufl.corealg.traversal import traverse_unique_terminals
from ufl.sobolevspace import H1

# Export list for ufl.classes
__all_classes__ = ["AbstractDomain", "Mesh", "MeshView"]


class AbstractDomain(object):
    """Symbolic representation of a geometric domain.

    Domain has only a geometric and a topological dimension.
    """

    def __init__(self, topological_dimension, geometric_dimension):
        """Initialise."""
        # Validate dimensions
        if not isinstance(geometric_dimension, numbers.Integral):
            raise ValueError(
                f"Expecting integer geometric dimension, not {geometric_dimension.__class__}"
            )
        if not isinstance(topological_dimension, numbers.Integral):
            raise ValueError(
                f"Expecting integer topological dimension, not {topological_dimension.__class__}"
            )
        if topological_dimension > geometric_dimension:
            raise ValueError("Topological dimension cannot be larger than geometric dimension.")

        # Store validated dimensions
        self._topological_dimension = topological_dimension
        self._geometric_dimension = geometric_dimension

    def geometric_dimension(self):
        """Return the dimension of the space this domain is embedded in."""
        return self._geometric_dimension

    def topological_dimension(self):
        """Return the dimension of the topology of this domain."""
        return self._topological_dimension


# TODO: Would it be useful to have a domain representing R^d? E.g. for
# Expression.
# class EuclideanSpace(AbstractDomain):
#     def __init__(self, geometric_dimension):
#         AbstractDomain.__init__(self, geometric_dimension, geometric_dimension)


@attach_ufl_id
class Mesh(AbstractDomain, UFLObject):
    """Symbolic representation of a mesh."""

    def __init__(self, coordinate_element, ufl_id=None, cargo=None):
        """Initialise."""
        self._ufl_id = self._init_ufl_id(ufl_id)

        # Store reference to object that will not be used by UFL
        self._ufl_cargo = cargo
        if cargo is not None and cargo.ufl_id() != self._ufl_id:
            raise ValueError("Expecting cargo object (e.g. dolfin.Mesh) to have the same ufl_id.")

        # No longer accepting coordinates provided as a Coefficient
        from ufl.coefficient import Coefficient

        if isinstance(coordinate_element, (Coefficient, AbstractCell)):
            raise ValueError("Expecting a coordinate element in the ufl.Mesh construct.")

        # Store coordinate element
        self._ufl_coordinate_element = coordinate_element

        # Derive dimensions from element
        (gdim,) = coordinate_element.reference_value_shape
        tdim = coordinate_element.cell.topological_dimension()
        AbstractDomain.__init__(self, tdim, gdim)

    def ufl_cargo(self):
        """Return carried object that will not be used by UFL."""
        return self._ufl_cargo

    def ufl_coordinate_element(self):
        """Get the coordinate element."""
        return self._ufl_coordinate_element

    def ufl_cell(self):
        """Get the cell."""
        return self._ufl_coordinate_element.cell

    def is_piecewise_linear_simplex_domain(self):
        """Check if the domain is a piecewise linear simplex."""
        ce = self._ufl_coordinate_element
        return ce.embedded_superdegree <= 1 and ce in H1 and self.ufl_cell().is_simplex()

    def __repr__(self):
        """Representation."""
        r = "Mesh(%s, %s)" % (repr(self._ufl_coordinate_element), repr(self._ufl_id))
        return r

    def __str__(self):
        """Format as a string."""
        return "<Mesh #%s>" % (self._ufl_id,)

    def _ufl_hash_data_(self):
        """UFL hash data."""
        return (self._ufl_id, self._ufl_coordinate_element)

    def _ufl_signature_data_(self, renumbering):
        """UFL signature data."""
        return ("Mesh", renumbering[self], self._ufl_coordinate_element)

    # NB! Dropped __lt__ here, don't want users to write 'mesh1 <
    # mesh2'.
    def _ufl_sort_key_(self):
        """UFL sort key."""
        typespecific = (self._ufl_id, self._ufl_coordinate_element)
        return (self.geometric_dimension(), self.topological_dimension(), "Mesh", typespecific)


@attach_ufl_id
class MeshView(AbstractDomain, UFLObject):
    """Symbolic representation of a mesh."""

    def __init__(self, mesh, topological_dimension, ufl_id=None):
        """Initialise."""
        self._ufl_id = self._init_ufl_id(ufl_id)

        # Store mesh
        self._ufl_mesh = mesh

        # Derive dimensions from element
        coordinate_element = mesh.ufl_coordinate_element()
        (gdim,) = coordinate_element.value_shape
        tdim = coordinate_element.cell.topological_dimension()
        AbstractDomain.__init__(self, tdim, gdim)

    def ufl_mesh(self):
        """Get the mesh."""
        return self._ufl_mesh

    def ufl_cell(self):
        """Get the cell."""
        return self._ufl_mesh.ufl_cell()

    def is_piecewise_linear_simplex_domain(self):
        """Check if the domain is a piecewise linear simplex."""
        return self._ufl_mesh.is_piecewise_linear_simplex_domain()

    def __repr__(self):
        """Representation."""
        tdim = self.topological_dimension()
        r = "MeshView(%s, %s, %s)" % (repr(self._ufl_mesh), repr(tdim), repr(self._ufl_id))
        return r

    def __str__(self):
        """Format as a string."""
        return "<MeshView #%s of dimension %d over mesh %s>" % (
            self._ufl_id,
            self.topological_dimension(),
            self._ufl_mesh,
        )

    def _ufl_hash_data_(self):
        """UFL hash data."""
        return (self._ufl_id,) + self._ufl_mesh._ufl_hash_data_()

    def _ufl_signature_data_(self, renumbering):
        """UFL signature data."""
        return ("MeshView", renumbering[self], self._ufl_mesh._ufl_signature_data_(renumbering))

    # NB! Dropped __lt__ here, don't want users to write 'mesh1 <
    # mesh2'.
    def _ufl_sort_key_(self):
        """UFL sort key."""
        typespecific = (self._ufl_id, self._ufl_mesh)
        return (self.geometric_dimension(), self.topological_dimension(), "MeshView", typespecific)


def as_domain(domain):
    """Convert any valid object to an AbstractDomain type."""
    if isinstance(domain, AbstractDomain):
        # Modern UFL files and dolfin behaviour
        return domain

    try:
        return extract_unique_domain(domain)
    except AttributeError:
        return domain.ufl_domain()


def sort_domains(domains):
    """Sort domains in a canonical ordering."""
    return tuple(sorted(domains, key=lambda domain: domain._ufl_sort_key_()))


def join_domains(domains):
    """Take a list of domains and return a tuple with only unique domain objects.

    Checks that domains with the same id are compatible.
    """
    # Use hashing to join domains, ignore None
    domains = set(domains) - set((None,))
    if not domains:
        return ()

    # Check geometric dimension compatibility
    gdims = set()
    for domain in domains:
        gdims.add(domain.geometric_dimension())
    if len(gdims) != 1:
        raise ValueError("Found domains with different geometric dimensions.")

    return domains


# TODO: Move these to an analysis module?


def extract_domains(expr):
    """Return all domains expression is defined on."""
    domainlist = []
    for t in traverse_unique_terminals(expr):
        domainlist.extend(t.ufl_domains())
    return sorted(
        join_domains(domainlist),
        key=lambda D: (D.topological_dimension(), D.ufl_cell(), D.ufl_id()),
    )


def extract_unique_domain(expr):
    """Return the single unique domain expression is defined on or throw an error."""
    domains = extract_domains(expr)
    if len(domains) == 1:
        return domains[0]
    elif domains:
        raise ValueError("Found multiple domains, cannot return just one.")
    else:
        return None


def find_geometric_dimension(expr):
    """Find the geometric dimension of an expression."""
    gdims = set()
    for t in traverse_unique_terminals(expr):
        domain = extract_unique_domain(t)
        if domain is not None:
            gdims.add(domain.geometric_dimension())

    if len(gdims) != 1:
        raise ValueError("Cannot determine geometric dimension from expression.")
    (gdim,) = gdims
    return gdim



"""Non-recursive traversal-based hash computation algorithm.

Fast iteration over nodes in an ``Expr`` DAG to compute
memoized hashes for all unique nodes.
"""

# Copyright (C) 2015 Martin Sandve Alnæs
#
# This file is part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
#
# Modified by Massimiliano Leoni, 2016


def compute_expr_hash(expr):
    """Compute hashes of *expr* and all its nodes efficiently, without using Python recursion."""
    if expr._hash is not None:
        return expr._hash
    # Postorder traversal, can't use unique_post_traversal, since that
    # uses a set which requires that this hash is computed.
    lifo = [(expr, list(expr.ufl_operands))]
    while lifo:
        expr, deps = lifo[-1]
        for i, dep in enumerate(deps):
            if dep is not None and dep._hash is None:
                lifo.append((dep, list(dep.ufl_operands)))
                deps[i] = None
                break
        else:
            if expr._hash is None:
                expr._hash = expr._ufl_compute_hash_()
            lifo.pop()
    return expr._hash



"""Signature computation for forms."""
# Copyright (C) 2012-2016 Martin Sandve Alnæs
#
# This file is part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later

import hashlib

from ufl.algorithms.domain_analysis import canonicalize_metadata
from ufl.classes import (
    Argument,
    Coefficient,
    Constant,
    ConstantValue,
    ExprList,
    ExprMapping,
    GeometricQuantity,
    Index,
    Label,
    MultiIndex,
)
from ufl.corealg.traversal import traverse_unique_terminals, unique_post_traversal


def compute_multiindex_hashdata(expr, index_numbering):
    """Compute multiindex hashdata."""
    data = []
    for i in expr:
        if isinstance(i, Index):
            j = index_numbering.get(i)
            if j is None:
                # Use negative ints for Index
                j = -(len(index_numbering) + 1)
                index_numbering[i] = j
            data.append(j)
        else:
            # Use nonnegative ints for FixedIndex
            data.append(int(i))
    return tuple(data)


def compute_terminal_hashdata(expressions, renumbering):
    """Compute terminal hashdata."""
    if not isinstance(expressions, list):
        expressions = [expressions]
    assert renumbering is not None

    # Extract a unique numbering of free indices, as well as form
    # arguments, and just take repr of the rest of the terminals while
    # we're iterating over them
    terminal_hashdata = {}
    index_numbering = {}
    for expression in expressions:
        for expr in traverse_unique_terminals(expression):
            if isinstance(expr, MultiIndex):
                # Indices need a canonical numbering for a stable
                # signature, thus this algorithm
                data = compute_multiindex_hashdata(expr, index_numbering)

            elif isinstance(expr, ConstantValue):
                data = expr._ufl_signature_data_(renumbering)

            elif isinstance(expr, Coefficient):
                data = expr._ufl_signature_data_(renumbering)

            elif isinstance(expr, Constant):
                data = expr._ufl_signature_data_(renumbering)

            elif isinstance(expr, Argument):
                data = expr._ufl_signature_data_(renumbering)

            elif isinstance(expr, GeometricQuantity):
                data = expr._ufl_signature_data_(renumbering)

            elif isinstance(expr, Label):
                data = expr._ufl_signature_data_(renumbering)

            elif isinstance(expr, ExprList):
                # Not really a terminal but can have 0 operands...
                data = "[]"

            elif isinstance(expr, ExprMapping):
                # Not really a terminal but can have 0 operands...
                data = "{}"

            else:
                raise ValueError(f"Unknown terminal type {type(expr)}")

            terminal_hashdata[expr] = data

    return terminal_hashdata


def compute_expression_hashdata(expression, terminal_hashdata) -> bytes:
    """Compute expression hashdata."""
    cache = {}

    for expr in unique_post_traversal(expression):
        # Uniquely traverse tree and hash each node
        # E.g. (a + b*c) is hashed as hash([+, hash(a), hash([*, hash(b), hash(c)])])
        # Traversal uses post pattern, so children hashes are cached
        if expr._ufl_is_terminal_:
            data = [terminal_hashdata[expr]]
        else:
            data = [expr._ufl_typecode_]

            for op in expr.ufl_operands:
                data += [cache[op]]
        cache[expr] = hashlib.sha512(str(data).encode("utf-8")).digest()
    return cache[expression]


def compute_expression_signature(expr, renumbering):  # FIXME: Fix callers
    """Compute expression signature."""
    # FIXME: Rewrite in terms of compute_form_signature?

    # Build hashdata for all terminals first
    terminal_hashdata = compute_terminal_hashdata([expr], renumbering)

    # Build hashdata for full expression
    expression_hashdata = compute_expression_hashdata(expr, terminal_hashdata)

    # Pass it through a seriously overkill hashing algorithm
    # (should we use sha1 instead?)
    return expression_hashdata.hex()


def compute_form_signature(form, renumbering):  # FIXME: Fix callers
    """Compute form signature."""
    # Extract integrands
    integrals = form.integrals()
    integrands = [integral.integrand() for integral in integrals]

    # Build hashdata for all terminals first, with on-the-fly
    # replacement of functions and index labels.
    terminal_hashdata = compute_terminal_hashdata(integrands, renumbering)

    # Build hashdata for each integral
    hashdata = []
    for integral in integrals:
        # Compute hash data for expression, this is the expensive part
        integrand_hashdata = compute_expression_hashdata(integral.integrand(), terminal_hashdata)

        domain_hashdata = integral.ufl_domain()._ufl_signature_data_(renumbering)

        # Collect all data about integral that should be reflected in
        # signature, including compiler data but not domain data,
        # because compiler data affects the way the integral is
        # compiled while domain data is only carried for convenience
        # in the problem solving environment.
        integral_hashdata = (
            integrand_hashdata,
            domain_hashdata,
            integral.integral_type(),
            integral.subdomain_id(),
            canonicalize_metadata(integral.metadata()),
        )

        hashdata.append(integral_hashdata)

    # Pass it through a seriously overkill hashing algorithm
    # (should we use sha1 instead?)
    data = str(hashdata).encode("utf-8")
    return hashlib.sha512(data).hexdigest()



#
# Author: Martin Sandve Alnes
# Date: 2008-10-30
#
from ufl import (
    Coefficient,
    FunctionSpace,
    Mesh,
    action,
    adjoint,
    derivative,
    dx,
    grad,
    inner,
    triangle,
)
from ufl.finiteelement import FiniteElement
from ufl.pullback import identity_pullback
from ufl.sobolevspace import H1

element = FiniteElement("Lagrange", triangle, 1, (), identity_pullback, H1)
domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
space = FunctionSpace(domain, element)

w = Coefficient(space)

# H1 semi-norm
f = inner(grad(w), grad(w)) / 2 * dx
# grad(w) : grad(v)
b = derivative(f, w)
# stiffness matrix, grad(u) : grad(v)
a = derivative(b, w)

# adjoint, grad(v) : grad(u)
astar = adjoint(a)
# action of adjoint, grad(v) : grad(w)
astaraction = action(astar)

forms = [f, b, a, astar, astaraction]



"""Tests of the change to local representaiton algorithms."""

from ufl import Coefficient, FunctionSpace, Mesh, as_tensor, grad, indices, triangle
from ufl.algorithms import change_to_reference_grad
from ufl.algorithms.renumbering import renumber_indices
from ufl.classes import JacobianInverse, ReferenceGrad
from ufl.finiteelement import FiniteElement
from ufl.pullback import identity_pullback
from ufl.sobolevspace import H1


def test_change_to_reference_grad():
    cell = triangle
    domain = Mesh(FiniteElement("Lagrange", cell, 1, (2,), identity_pullback, H1))
    U = FunctionSpace(domain, FiniteElement("Lagrange", cell, 1, (), identity_pullback, H1))
    V = FunctionSpace(domain, FiniteElement("Lagrange", cell, 1, (2,), identity_pullback, H1))
    u = Coefficient(U)
    v = Coefficient(V)
    Jinv = JacobianInverse(domain)
    i, j, k = indices(3)
    q, r, s = indices(3)
    (t,) = indices(1)

    # Single grad change on a scalar function
    expr = grad(u)
    actual = change_to_reference_grad(expr)
    expected = as_tensor(Jinv[k, i] * ReferenceGrad(u)[k], (i,))
    assert renumber_indices(actual) == renumber_indices(expected)

    # Single grad change on a vector valued function
    expr = grad(v)
    actual = change_to_reference_grad(expr)
    expected = as_tensor(Jinv[k, j] * ReferenceGrad(v)[i, k], (i, j))
    assert renumber_indices(actual) == renumber_indices(expected)

    # Multiple grads should work fine for affine domains:
    expr = grad(grad(u))
    actual = change_to_reference_grad(expr)
    expected = as_tensor(Jinv[s, j] * (Jinv[r, i] * ReferenceGrad(ReferenceGrad(u))[r, s]), (i, j))
    assert renumber_indices(actual) == renumber_indices(expected)

    expr = grad(grad(grad(u)))
    actual = change_to_reference_grad(expr)
    expected = as_tensor(
        Jinv[s, k]
        * (Jinv[r, j] * (Jinv[q, i] * ReferenceGrad(ReferenceGrad(ReferenceGrad(u)))[q, r, s])),
        (i, j, k),
    )
    assert renumber_indices(actual) == renumber_indices(expected)

    # Multiple grads on a vector valued function
    expr = grad(grad(v))
    actual = change_to_reference_grad(expr)
    expected = as_tensor(
        Jinv[s, j] * (Jinv[r, i] * ReferenceGrad(ReferenceGrad(v))[t, r, s]), (t, i, j)
    )
    assert renumber_indices(actual) == renumber_indices(expected)

    expr = grad(grad(grad(v)))
    actual = change_to_reference_grad(expr)
    expected = as_tensor(
        Jinv[s, k]
        * (Jinv[r, j] * (Jinv[q, i] * ReferenceGrad(ReferenceGrad(ReferenceGrad(v)))[t, q, r, s])),
        (t, i, j, k),
    )
    assert renumber_indices(actual) == renumber_indices(expected)

    # print tree_format(expected)
    # print tree_format(actual)
    # print tree_format(renumber_indices(actual))
    # print tree_format(renumber_indices(expected))



"""Tests of the various ways Measure objects can be created and used."""

from mockobjects import MockMesh, MockMeshFunction

from ufl import Cell, Coefficient, FunctionSpace, Measure, Mesh, as_ufl, dC, dI, dO, triangle
from ufl.finiteelement import FiniteElement
from ufl.pullback import identity_pullback
from ufl.sobolevspace import H1


def test_construct_forms_from_default_measures():
    # Create defaults:
    dx = Measure("dx")
    # dE = Measure("dE")

    ds = Measure("ds")
    dS = Measure("dS")

    dP = Measure("dP")
    # dV = Measure("dV")

    dc = Measure("dc")
    # dC = Measure("dC")
    # dO = Measure("dO")
    # dI = Measure("dI")

    ds_b = Measure("ds_b")
    ds_t = Measure("ds_t")
    ds_v = Measure("ds_v")
    dS_h = Measure("dS_h")
    dS_v = Measure("dS_v")

    # Check that names are mapped properly
    assert dx.integral_type() == "cell"
    # assert dE.integral_type() == "macro_cell"

    assert ds.integral_type() == "exterior_facet"
    assert dS.integral_type() == "interior_facet"

    assert dP.integral_type() == "vertex"
    # TODO: Change dP to dV:
    # assert dP.integral_type() == "point"
    # assert dV.integral_type() == "vertex"

    assert dc.integral_type() == "custom"
    assert dC.integral_type() == "cutcell"
    assert dO.integral_type() == "overlap"
    assert dI.integral_type() == "interface"

    # TODO: Remove firedrake hacks:
    assert ds_b.integral_type() == "exterior_facet_bottom"
    assert ds_t.integral_type() == "exterior_facet_top"
    assert ds_v.integral_type() == "exterior_facet_vert"
    assert dS_h.integral_type() == "interior_facet_horiz"
    assert dS_v.integral_type() == "interior_facet_vert"

    # Check that defaults are set properly
    assert dx.ufl_domain() is None
    assert dx.metadata() == {}

    # Check that we can create a basic form with default measure
    one = as_ufl(1)
    one * dx(Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1)))


def test_foo():
    # Define a manifold domain, allows checking gdim/tdim mixup errors
    gdim = 3
    tdim = 2
    cell = Cell("triangle")
    mymesh = MockMesh(9)
    mydomain = Mesh(
        FiniteElement("Lagrange", cell, 1, (gdim,), identity_pullback, H1), ufl_id=9, cargo=mymesh
    )

    assert cell.topological_dimension() == tdim
    assert cell.cellname() == "triangle"
    assert mydomain.topological_dimension() == tdim
    assert mydomain.geometric_dimension() == gdim
    assert mydomain.ufl_cell() == cell
    assert mydomain.ufl_id() == 9
    assert mydomain.ufl_cargo() == mymesh

    # Define a coefficient for use in tests below
    V = FunctionSpace(mydomain, FiniteElement("Lagrange", cell, 1, (), identity_pullback, H1))
    f = Coefficient(V)

    # Test definition of a custom measure with explicit parameters
    metadata = {"opt": True}
    mydx = Measure("dx", domain=mydomain, subdomain_id=3, metadata=metadata)
    assert mydx.ufl_domain().ufl_id() == mydomain.ufl_id()
    assert mydx.metadata() == metadata
    M = f * mydx

    # Compatibility:
    dx = Measure("dx")
    # domain=None,
    # subdomain_id="everywhere",
    # metadata=None)
    assert dx.ufl_domain() is None
    assert dx.subdomain_id() == "everywhere"

    # Set subdomain_id to "everywhere", still no domain set
    dxe = dx()
    assert dxe.ufl_domain() is None
    assert dxe.subdomain_id() == "everywhere"

    # Set subdomain_id to 5, still no domain set
    dx5 = dx(5)
    assert dx5.ufl_domain() is None
    assert dx5.subdomain_id() == 5

    # Check that original dx is untouched
    assert dx.ufl_domain() is None
    assert dx.subdomain_id() == "everywhere"

    # Set subdomain_id to (2,3), still no domain set
    dx23 = dx((2, 3))
    assert dx23.ufl_domain() is None
    assert dx23.subdomain_id(), 2 == 3

    # Map metadata to metadata, ffc interprets as before
    dxm = dx(metadata={"dummy": 123})
    # assert dxm.metadata() == {"dummy":123}
    assert dxm.metadata() == {"dummy": 123}  # Deprecated, TODO: Remove

    assert dxm.ufl_domain() is None
    assert dxm.subdomain_id() == "everywhere"

    # dxm = dx(metadata={"dummy":123})
    # assert dxm.metadata() == {"dummy":123}
    dxm = dx(metadata={"dummy": 123})
    assert dxm.metadata() == {"dummy": 123}

    assert dxm.ufl_domain() is None
    assert dxm.subdomain_id() == "everywhere"

    # Mock some dolfin data structures
    dx = Measure("dx")
    ds = Measure("ds")
    dS = Measure("dS")
    mesh = MockMesh(8)
    cell_domains = MockMeshFunction(1, mesh)
    exterior_facet_domains = MockMeshFunction(2, mesh)
    interior_facet_domains = MockMeshFunction(3, mesh)

    dxd = dx(subdomain_data=cell_domains)
    dsd = ds(subdomain_data=exterior_facet_domains)
    dSd = dS(subdomain_data=interior_facet_domains)
    # Current behaviour: no domain created, measure domain data is a single
    # object not a full dict
    assert dxd.ufl_domain() is None
    assert dsd.ufl_domain() is None
    assert dSd.ufl_domain() is None
    assert dxd.subdomain_data() is cell_domains
    assert dsd.subdomain_data() is exterior_facet_domains
    assert dSd.subdomain_data() is interior_facet_domains

    # Create some forms with these measures (used in checks below):
    Mx = f * dxd
    Ms = f**2 * dsd
    MS = f("+") * dSd
    M = f * dxd + f**2 * dsd + f("+") * dSd

    # Test extracting domain data from a form for each measure:
    (domain,) = Mx.ufl_domains()
    assert domain.ufl_id() == mydomain.ufl_id()
    assert domain.ufl_cargo() == mymesh
    assert len(Mx.subdomain_data()[mydomain]["cell"]) == 1
    assert Mx.subdomain_data()[mydomain]["cell"][0] == cell_domains

    (domain,) = Ms.ufl_domains()
    assert domain.ufl_cargo() == mymesh
    assert len(Ms.subdomain_data()[mydomain]["exterior_facet"]) == 1
    assert Ms.subdomain_data()[mydomain]["exterior_facet"][0] == exterior_facet_domains

    (domain,) = MS.ufl_domains()
    assert domain.ufl_cargo() == mymesh
    assert len(MS.subdomain_data()[mydomain]["interior_facet"]) == 1
    assert MS.subdomain_data()[mydomain]["interior_facet"][0] == interior_facet_domains

    # Test joining of these domains in a single form
    (domain,) = M.ufl_domains()
    assert domain.ufl_cargo() == mymesh
    assert len(M.subdomain_data()[mydomain]["cell"]) == 1
    assert M.subdomain_data()[mydomain]["cell"][0] == cell_domains
    assert len(M.subdomain_data()[mydomain]["exterior_facet"]) == 1
    assert M.subdomain_data()[mydomain]["exterior_facet"][0] == exterior_facet_domains
    assert len(M.subdomain_data()[mydomain]["interior_facet"]) == 1
    assert M.subdomain_data()[mydomain]["interior_facet"][0] == interior_facet_domains



import math

from ufl import (
    Coefficient,
    FunctionSpace,
    Mesh,
    TestFunction,
    TrialFunction,
    VectorConstant,
    acos,
    as_tensor,
    as_ufl,
    asin,
    atan,
    cos,
    cosh,
    dx,
    exp,
    i,
    j,
    ln,
    max_value,
    min_value,
    outer,
    sin,
    sinh,
    tan,
    tanh,
    triangle,
)
from ufl.algorithms import compute_form_data
from ufl.finiteelement import FiniteElement
from ufl.pullback import identity_pullback
from ufl.sobolevspace import H1


def xtest_zero_times_argument(self):
    # FIXME: Allow zero forms
    element = FiniteElement("Lagrange", triangle, 1, (), identity_pullback, H1)
    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    space = FunctionSpace(domain, element)
    v = TestFunction(space)
    u = TrialFunction(space)
    L = 0 * v * dx
    a = 0 * (u * v) * dx
    b = (0 * u) * v * dx
    assert len(compute_form_data(L).arguments) == 1
    assert len(compute_form_data(a).arguments) == 2
    assert len(compute_form_data(b).arguments) == 2


def test_divisions(self):
    element = FiniteElement("Lagrange", triangle, 1, (), identity_pullback, H1)
    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    space = FunctionSpace(domain, element)
    f = Coefficient(space)

    # Test simplification of division by 1
    a = f
    b = f / 1
    assert a == b

    # Test simplification of division by 1.0
    a = f
    b = f / 1.0
    assert a == b

    # Test simplification of division by of zero by something
    a = 0 / f
    b = 0 * f
    assert a == b

    # Test simplification of division by self (this simplification has been disabled)
    # a = f/f
    # b = 1
    # assert a == b


def test_products(self):
    element = FiniteElement("Lagrange", triangle, 1, (), identity_pullback, H1)
    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    space = FunctionSpace(domain, element)
    f = Coefficient(space)
    g = Coefficient(space)

    # Test simplification of literal multiplication
    assert f * 0 == as_ufl(0)
    assert 0 * f == as_ufl(0)
    assert 1 * f == f
    assert f * 1 == f
    assert as_ufl(2) * as_ufl(3) == as_ufl(6)
    assert as_ufl(2.0) * as_ufl(3.0) == as_ufl(6.0)

    # Test reordering of operands
    assert f * g == g * f

    # Test simplification of self-multiplication (this simplification has been disabled)
    # assert f*f == f**2


def test_sums(self):
    element = FiniteElement("Lagrange", triangle, 1, (), identity_pullback, H1)
    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    space = FunctionSpace(domain, element)
    f = Coefficient(space)
    g = Coefficient(space)

    # Test reordering of operands
    assert f + g == g + f

    # Test adding zero
    assert f + 0 == f
    assert 0 + f == f

    # Test collapsing of basic sum (this simplification has been disabled)
    # assert f + f == 2 * f

    # Test reordering of operands and collapsing sum
    a = f + g + f  # not collapsed, but ordered
    b = g + f + f  # not collapsed, but ordered
    c = (g + f) + f  # not collapsed, but ordered
    d = f + (f + g)  # not collapsed, but ordered
    assert a == b
    assert a == c
    assert a == d

    # Test reordering of operands and collapsing sum
    a = f + f + g  # collapsed
    b = g + (f + f)  # collapsed
    assert a == b


def test_mathfunctions(self):
    for a in (0.1, 0.3, 0.9):
        assert math.sin(a) == sin(a)
        assert math.cos(a) == cos(a)
        assert math.tan(a) == tan(a)
        assert math.sinh(a) == sinh(a)
        assert math.cosh(a) == cosh(a)
        assert math.tanh(a) == tanh(a)
        assert math.asin(a) == asin(a)
        assert math.acos(a) == acos(a)
        assert math.atan(a) == atan(a)
        assert math.exp(a) == exp(a)
        assert math.log(a) == ln(a)
        # TODO: Implement automatic simplification of conditionals?
        assert a == float(max_value(a, a - 1))
        # TODO: Implement automatic simplification of conditionals?
        assert a == float(min_value(a, a + 1))


def test_indexing(self):
    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    u = VectorConstant(domain)
    v = VectorConstant(domain)

    A = outer(u, v)
    A2 = as_tensor(A[i, j], (i, j))
    assert A2 == A

    Bij = u[i] * v[j]
    Bij2 = as_tensor(Bij, (i, j))[i, j]
    as_tensor(Bij, (i, j))
    assert Bij2 == Bij



"""Algorithm for expanding compound expressions."""

# Copyright (C) 2008-2016 Martin Sandve Alnæs and Anders Logg
#
# This file is part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
#
# Modified by Anders Logg, 2009-2010

from ufl.algorithms.map_integrands import map_integrand_dags
from ufl.classes import Conj, Grad, Product
from ufl.compound_expressions import cofactor_expr, determinant_expr, deviatoric_expr, inverse_expr
from ufl.core.multiindex import Index, indices
from ufl.corealg.multifunction import MultiFunction
from ufl.tensors import as_matrix, as_tensor, as_vector


class LowerCompoundAlgebra(MultiFunction):
    """Expands high level compound operators to equivalent representations using basic operators."""

    def __init__(self):
        """Initialize."""
        MultiFunction.__init__(self)

    ufl_type = MultiFunction.reuse_if_untouched

    # ------------ Compound tensor operators

    def trace(self, o, A):
        """Lower a trace."""
        i = Index()
        return A[i, i]

    def transposed(self, o, A):
        """Lower a transposed."""
        i, j = indices(2)
        return as_tensor(A[i, j], (j, i))

    def deviatoric(self, o, A):
        """Lower a deviatoric."""
        return deviatoric_expr(A)

    def skew(self, o, A):
        """Lower a skew."""
        i, j = indices(2)
        return as_matrix((A[i, j] - A[j, i]) / 2, (i, j))

    def sym(self, o, A):
        """Lower a sym."""
        i, j = indices(2)
        return as_matrix((A[i, j] + A[j, i]) / 2, (i, j))

    def cross(self, o, a, b):
        """Lower a cross."""

        def c(i, j):
            return Product(a[i], b[j]) - Product(a[j], b[i])

        return as_vector((c(1, 2), c(2, 0), c(0, 1)))

    def perp(self, o, a):
        """Lower a perp."""
        return as_vector([-a[1], a[0]])

    def dot(self, o, a, b):
        """Lower a dot."""
        ai = indices(len(a.ufl_shape) - 1)
        bi = indices(len(b.ufl_shape) - 1)
        k = (Index(),)
        # Creates a single IndexSum over a Product
        s = a[ai + k] * b[k + bi]
        return as_tensor(s, ai + bi)

    def inner(self, o, a, b):
        """Lower an inner."""
        ash = a.ufl_shape
        bsh = b.ufl_shape
        if ash != bsh:
            raise ValueError("Nonmatching shapes.")
        ii = indices(len(ash))
        # Creates multiple IndexSums over a Product
        s = a[ii] * Conj(b[ii])
        return s

    def outer(self, o, a, b):
        """Lower an outer."""
        ii = indices(len(a.ufl_shape))
        jj = indices(len(b.ufl_shape))
        # Create a Product with no shared indices
        s = Conj(a[ii]) * b[jj]
        return as_tensor(s, ii + jj)

    def determinant(self, o, A):
        """Lower a determinant."""
        return determinant_expr(A)

    def cofactor(self, o, A):
        """Lower a cofactor."""
        return cofactor_expr(A)

    def inverse(self, o, A):
        """Lower an inverse."""
        return inverse_expr(A)

    # ------------ Compound differential operators

    def div(self, o, a):
        """Lower a div."""
        i = Index()
        return a[..., i].dx(i)

    def nabla_div(self, o, a):
        """Lower a nabla_div."""
        i = Index()
        return a[i, ...].dx(i)

    def nabla_grad(self, o, a):
        """Lower a nabla_grad."""
        sh = a.ufl_shape
        if sh == ():
            return Grad(a)
        else:
            j = Index()
            ii = tuple(indices(len(sh)))
            return as_tensor(a[ii].dx(j), (j,) + ii)

    def curl(self, o, a):
        """Lower a curl."""

        # o = curl a = "[a.dx(1), -a.dx(0)]"            if a.ufl_shape == ()
        # o = curl a = "cross(nabla, (a0, a1, 0))[2]" if a.ufl_shape == (2,)
        # o = curl a = "cross(nabla, a)"              if a.ufl_shape == (3,)
        def c(i, j):
            """A component of curl."""
            return a[j].dx(i) - a[i].dx(j)

        sh = a.ufl_shape
        if sh == ():
            return as_vector((a.dx(1), -a.dx(0)))
        if sh == (2,):
            return c(0, 1)
        if sh == (3,):
            return as_vector((c(1, 2), c(2, 0), c(0, 1)))
        raise ValueError(f"Invalid shape {sh} of curl argument.")


def apply_algebra_lowering(expr):
    """Expands high level compound operators to equivalent representations using basic operators."""
    return map_integrand_dags(LowerCompoundAlgebra(), expr)



"""This module defines the Matrix class."""
# Copyright (C) 2021 India Marsden
#
# This file is part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
#
# Modified by Nacime Bouziani, 2021-2022.

from ufl.argument import Argument
from ufl.core.ufl_type import ufl_type
from ufl.form import BaseForm
from ufl.functionspace import AbstractFunctionSpace
from ufl.utils.counted import Counted

# --- The Matrix class represents a matrix, an assembled two form ---


@ufl_type()
class Matrix(BaseForm, Counted):
    """An assemble linear operator between two function spaces."""

    __slots__ = (
        "_count",
        "_counted_class",
        "_ufl_function_spaces",
        "ufl_operands",
        "_repr",
        "_hash",
        "_ufl_shape",
        "_arguments",
        "_coefficients",
        "_domains",
    )

    def __getnewargs__(self):
        """Get new args."""
        return (self._ufl_function_spaces[0], self._ufl_function_spaces[1], self._count)

    def __init__(self, row_space, column_space, count=None):
        """Initialise."""
        BaseForm.__init__(self)
        Counted.__init__(self, count, Matrix)

        if not isinstance(row_space, AbstractFunctionSpace):
            raise ValueError("Expecting a FunctionSpace as the row space.")

        if not isinstance(column_space, AbstractFunctionSpace):
            raise ValueError("Expecting a FunctionSpace as the column space.")

        self._ufl_function_spaces = (row_space, column_space)

        self.ufl_operands = ()
        self._domains = None
        self._hash = None
        self._repr = (
            f"Matrix({self._ufl_function_spaces[0]!r} "
            f"{self._ufl_function_spaces[1]!r}, {self._count!r})"
        )

    def ufl_function_spaces(self):
        """Get the tuple of function spaces of this coefficient."""
        return self._ufl_function_spaces

    def _analyze_form_arguments(self):
        """Define arguments of a matrix when considered as a form."""
        self._arguments = (
            Argument(self._ufl_function_spaces[0], 0),
            Argument(self._ufl_function_spaces[1], 1),
        )
        self._coefficients = ()

    def _analyze_domains(self):
        """Analyze which domains can be found in a Matrix."""
        from ufl.domain import join_domains

        # Collect unique domains
        self._domains = join_domains([fs.ufl_domain() for fs in self._ufl_function_spaces])

    def __str__(self):
        """Format as a string."""
        count = str(self._count)
        if len(count) == 1:
            return f"A_{count}"
        else:
            return f"A_{{{count}}}"

    def __repr__(self):
        """Representation."""
        return self._repr

    def __hash__(self):
        """Hash."""
        if self._hash is None:
            self._hash = hash(self._repr)
        return self._hash

    def equals(self, other):
        """Check equality."""
        if type(other) is not Matrix:
            return False
        if self is other:
            return True
        return (
            self._count == other._count and self._ufl_function_spaces == other._ufl_function_spaces
        )



"""Test the is_cellwise_constant function on all relevant terminal types."""

import pytest

from ufl import (
    Cell,
    CellDiameter,
    CellVolume,
    Circumradius,
    Coefficient,
    Constant,
    FacetArea,
    FacetNormal,
    FunctionSpace,
    Jacobian,
    JacobianDeterminant,
    JacobianInverse,
    MaxFacetEdgeLength,
    Mesh,
    MinFacetEdgeLength,
    SpatialCoordinate,
    TestFunction,
    hexahedron,
    interval,
    quadrilateral,
    tetrahedron,
    triangle,
)
from ufl.checks import is_cellwise_constant
from ufl.classes import (
    CellCoordinate,
    FacetJacobian,
    FacetJacobianDeterminant,
    FacetJacobianInverse,
)
from ufl.finiteelement import FiniteElement
from ufl.pullback import identity_pullback
from ufl.sobolevspace import H1, L2, HInf


def get_domains():
    all_cells = [
        # vertex,
        interval,
        triangle,
        quadrilateral,
        tetrahedron,
        hexahedron,
    ]
    return [
        Mesh(
            FiniteElement(
                "Lagrange", cell, 1, (cell.topological_dimension(),), identity_pullback, H1
            )
        )
        for cell in all_cells
    ]


def get_nonlinear():
    domains_with_quadratic_coordinates = []
    for D in get_domains():
        V = FiniteElement(
            "Lagrange",
            D.ufl_cell(),
            2,
            (D.ufl_cell().topological_dimension(),),
            identity_pullback,
            H1,
        )
        E = Mesh(V)
        domains_with_quadratic_coordinates.append(E)

    return domains_with_quadratic_coordinates


@pytest.fixture(params=list(range(5)))
def nonlinear_domains(request):
    return get_nonlinear()[request.param]


@pytest.fixture(params=list(range(10)))
def domains_not_linear(request):
    all_domains_not_linear = get_domains() + get_nonlinear()
    return all_domains_not_linear[request.param]


@pytest.fixture(params=list(range(15)))
def domains(request):
    domains = get_domains()
    domains_with_linear_coordinates = []
    for D in domains:
        V = FiniteElement(
            "Lagrange",
            D.ufl_cell(),
            1,
            (D.ufl_cell().topological_dimension(),),
            identity_pullback,
            H1,
        )
        E = Mesh(V)
        domains_with_linear_coordinates.append(E)

    all_domains = domains + domains_with_linear_coordinates + get_nonlinear()
    return all_domains[request.param]


@pytest.fixture(params=list(range(6)))
def affine_domains(request):
    affine_cells = [
        interval,
        triangle,
        tetrahedron,
    ]
    affine_domains = [
        Mesh(
            FiniteElement(
                "Lagrange", cell, 1, (cell.topological_dimension(),), identity_pullback, H1
            )
        )
        for cell in affine_cells
    ]

    affine_domains_with_linear_coordinates = []
    for D in affine_domains:
        V = FiniteElement(
            "Lagrange",
            D.ufl_cell(),
            1,
            (D.ufl_cell().topological_dimension(),),
            identity_pullback,
            H1,
        )
        E = Mesh(V)
        affine_domains_with_linear_coordinates.append(E)

    all_affine_domains = affine_domains + affine_domains_with_linear_coordinates
    return all_affine_domains[request.param]


@pytest.fixture(params=list(range(8)))
def affine_facet_domains(request):
    affine_facet_cells = [
        interval,
        triangle,
        quadrilateral,
        tetrahedron,
    ]
    affine_facet_domains = [
        Mesh(
            FiniteElement(
                "Lagrange", cell, 1, (cell.topological_dimension(),), identity_pullback, H1
            )
        )
        for cell in affine_facet_cells
    ]
    affine_facet_domains_with_linear_coordinates = []
    for D in affine_facet_domains:
        V = FiniteElement(
            "Lagrange",
            D.ufl_cell(),
            1,
            (D.ufl_cell().topological_dimension(),),
            identity_pullback,
            H1,
        )
        E = Mesh(V)
        affine_facet_domains_with_linear_coordinates.append(E)

    all_affine_facet_domains = affine_facet_domains + affine_facet_domains_with_linear_coordinates

    return all_affine_facet_domains[request.param]


@pytest.fixture(params=list(range(4)))
def nonaffine_domains(request):
    nonaffine_cells = [
        quadrilateral,
        hexahedron,
    ]
    nonaffine_domains = [
        Mesh(
            FiniteElement(
                "Lagrange", cell, 1, (cell.topological_dimension(),), identity_pullback, H1
            )
        )
        for cell in nonaffine_cells
    ]
    nonaffine_domains_with_linear_coordinates = []
    for D in nonaffine_domains:
        V = FiniteElement(
            "Lagrange",
            D.ufl_cell(),
            1,
            (D.ufl_cell().topological_dimension(),),
            identity_pullback,
            H1,
        )
        E = Mesh(V)
        nonaffine_domains_with_linear_coordinates.append(E)

    all_nonaffine_domains = nonaffine_domains + nonaffine_domains_with_linear_coordinates

    return all_nonaffine_domains[request.param]


@pytest.fixture(params=list(range(2)))
def nonaffine_facet_domains(request):
    nonaffine_facet_cells = [
        hexahedron,
    ]
    nonaffine_facet_domains = [
        Mesh(
            FiniteElement(
                "Lagrange", cell, 1, (cell.topological_dimension(),), identity_pullback, H1
            )
        )
        for cell in nonaffine_facet_cells
    ]
    nonaffine_facet_domains_with_linear_coordinates = []
    for D in nonaffine_facet_domains:
        V = FiniteElement(
            "Lagrange",
            D.ufl_cell(),
            1,
            (D.ufl_cell().topological_dimension(),),
            identity_pullback,
            H1,
        )
        E = Mesh(V)
        nonaffine_facet_domains_with_linear_coordinates.append(E)

    all_nonaffine_facet_domains = (
        nonaffine_facet_domains + nonaffine_facet_domains_with_linear_coordinates
    )

    return all_nonaffine_facet_domains[request.param]


def test_always_cellwise_constant_geometric_quantities(domains):
    "Test geometric quantities that are always constant over a cell."
    e = CellVolume(domains)
    assert is_cellwise_constant(e)
    e = CellDiameter(domains)
    assert is_cellwise_constant(e)
    e = Circumradius(domains)
    assert is_cellwise_constant(e)
    e = FacetArea(domains)
    assert is_cellwise_constant(e)
    e = MinFacetEdgeLength(domains)
    assert is_cellwise_constant(e)
    e = MaxFacetEdgeLength(domains)
    assert is_cellwise_constant(e)


def test_coordinates_never_cellwise_constant(domains):
    e = SpatialCoordinate(domains)
    assert not is_cellwise_constant(e)
    e = CellCoordinate(domains)
    assert not is_cellwise_constant(e)


def test_coordinates_never_cellwise_constant_vertex():
    # The only exception here:
    domains = Mesh(FiniteElement("Lagrange", Cell("vertex"), 1, (3,), identity_pullback, H1))
    assert domains.ufl_cell().cellname() == "vertex"
    e = SpatialCoordinate(domains)
    assert is_cellwise_constant(e)
    e = CellCoordinate(domains)
    assert is_cellwise_constant(e)


def mappings_are_cellwise_constant(domain, test):
    e = Jacobian(domain)
    assert is_cellwise_constant(e) == test
    e = JacobianDeterminant(domain)
    assert is_cellwise_constant(e) == test
    e = JacobianInverse(domain)
    assert is_cellwise_constant(e) == test
    if domain.topological_dimension() != 1:
        e = FacetJacobian(domain)
        assert is_cellwise_constant(e) == test
        e = FacetJacobianDeterminant(domain)
        assert is_cellwise_constant(e) == test
        e = FacetJacobianInverse(domain)
        assert is_cellwise_constant(e) == test


def test_mappings_are_cellwise_constant_on_linear_affine_cells(affine_domains):
    mappings_are_cellwise_constant(affine_domains, True)


def test_mappings_are_cellwise_not_constant_on_nonaffine_cells(nonaffine_domains):
    mappings_are_cellwise_constant(nonaffine_domains, False)


def test_mappings_are_cellwise_not_constant_on_nonlinear_cells(nonlinear_domains):
    mappings_are_cellwise_constant(nonlinear_domains, False)


def facetnormal_cellwise_constant(domain, test):
    e = FacetNormal(domain)
    assert is_cellwise_constant(e) == test


def test_facetnormal_cellwise_constant_affine(affine_facet_domains):
    facetnormal_cellwise_constant(affine_facet_domains, True)


def test_facetnormal_not_cellwise_constant_nonaffine(nonaffine_facet_domains):
    facetnormal_cellwise_constant(nonaffine_facet_domains, False)


def test_facetnormal_not_cellwise_constant_nonlinear(nonlinear_domains):
    facetnormal_cellwise_constant(nonlinear_domains, False)


def test_coefficient_sometimes_cellwise_constant(domains_not_linear):
    e = Constant(domains_not_linear)
    assert is_cellwise_constant(e)

    V = FiniteElement(
        "Discontinuous Lagrange", domains_not_linear.ufl_cell(), 0, (), identity_pullback, L2
    )
    d = domains_not_linear.ufl_cell().topological_dimension()
    domain = Mesh(
        FiniteElement("Lagrange", domains_not_linear.ufl_cell(), 1, (d,), identity_pullback, H1)
    )
    space = FunctionSpace(domain, V)
    e = Coefficient(space)
    assert is_cellwise_constant(e)

    V = FiniteElement("Real", domains_not_linear.ufl_cell(), 0, (), identity_pullback, HInf)
    space = FunctionSpace(domain, V)
    e = Coefficient(space)
    assert is_cellwise_constant(e)

    # This should be true, but that has to wait for a fix of issue #13
    # e = TestFunction(V)
    # assert is_cellwise_constant(e)
    # V = FiniteElement("R", domains_not_linear.ufl_cell(), 0)
    # e = TestFunction(V)
    # assert is_cellwise_constant(e)


def test_coefficient_mostly_not_cellwise_constant(domains_not_linear):
    V = FiniteElement(
        "Discontinuous Lagrange", domains_not_linear.ufl_cell(), 1, (), identity_pullback, L2
    )
    d = domains_not_linear.ufl_cell().topological_dimension()
    domain = Mesh(
        FiniteElement("Lagrange", domains_not_linear.ufl_cell(), 1, (d,), identity_pullback, H1)
    )
    space = FunctionSpace(domain, V)
    e = Coefficient(space)
    assert not is_cellwise_constant(e)
    e = TestFunction(space)
    assert not is_cellwise_constant(e)



# Copyright (C) 2012 Marie E. Rognes (meg@simula.no)
#
# This file is part of UFL.
#
# UFL is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# UFL is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with UFL. If not, see <http://www.gnu.org/licenses/>.
#
# First added:  2012-08-16
# Last changed: 2012-08-16
from ufl import (
    FunctionSpace,
    Mesh,
    TensorProductElement,
    TestFunction,
    TrialFunction,
    dx,
    interval,
    tetrahedron,
    triangle,
)
from ufl.finiteelement import FiniteElement
from ufl.pullback import identity_pullback
from ufl.sobolevspace import H1, L2

V0 = FiniteElement("Lagrange", triangle, 1, (), identity_pullback, H1)
V1 = FiniteElement("DG", interval, 0, (), identity_pullback, L2)
V2 = FiniteElement("DG", tetrahedron, 0, (), identity_pullback, L2)

V = TensorProductElement(V0, V1, V2)

c0 = FiniteElement("CG", triangle, 1)
c1 = FiniteElement("CG", interval, 1)
c2 = FiniteElement("CG", tetrahedron, 1)
domain = Mesh(TensorProductElement(c0, c1, c2))
space = FunctionSpace(domain, V)

u = TrialFunction(space)
v = TestFunction(space)

dxxx = dx * dx * dx
a = u * v * dxxx



"""Algorithm for replacing terminals in an expression."""

# Copyright (C) 2008-2016 Martin Sandve Alnæs and Anders Logg
#
# This file is part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
#
# Modified by Anders Logg, 2009-2010

from ufl.algorithms.analysis import has_exact_type
from ufl.algorithms.map_integrands import map_integrand_dags
from ufl.classes import BaseForm, CoefficientDerivative
from ufl.constantvalue import as_ufl
from ufl.core.external_operator import ExternalOperator
from ufl.core.interpolate import Interpolate
from ufl.corealg.multifunction import MultiFunction


class Replacer(MultiFunction):
    """Replacer."""

    def __init__(self, mapping):
        """Initialize."""
        super().__init__()
        self.mapping = mapping

        # One can replace Coarguments by 1-Forms
        def get_shape(x):
            """Get the shape of an object."""
            if isinstance(x, BaseForm):
                return x.arguments()[0].ufl_shape
            return x.ufl_shape

        if not all(get_shape(k) == get_shape(v) for k, v in mapping.items()):
            raise ValueError(
                "Replacement expressions must have the same shape as what they replace."
            )

    def ufl_type(self, o, *args):
        """Replace a ufl_type."""
        try:
            return self.mapping[o]
        except KeyError:
            return self.reuse_if_untouched(o, *args)

    def external_operator(self, o):
        """Replace an external_operator."""
        o = self.mapping.get(o) or o
        if isinstance(o, ExternalOperator):
            new_ops = tuple(replace(op, self.mapping) for op in o.ufl_operands)
            new_args = tuple(replace(arg, self.mapping) for arg in o.argument_slots())
            return o._ufl_expr_reconstruct_(*new_ops, argument_slots=new_args)
        return o

    def interpolate(self, o):
        """Replace an interpolate."""
        o = self.mapping.get(o) or o
        if isinstance(o, Interpolate):
            new_args = tuple(replace(arg, self.mapping) for arg in o.argument_slots())
            return o._ufl_expr_reconstruct_(*reversed(new_args))
        return o

    def coefficient_derivative(self, o):
        """Replace a coefficient derivative."""
        raise ValueError("Derivatives should be applied before executing replace.")


def replace(e, mapping):
    """Replace subexpressions in expression.

    Params:
        e: An Expr or Form
        mapping: A dict with from:to replacements to perform

    Returns:
        The expression with replacements performed
    """
    mapping2 = dict((k, as_ufl(v)) for (k, v) in mapping.items())

    # Workaround for problem with delayed derivative evaluation
    # The problem is that J = derivative(f(g, h), g) does not evaluate immediately
    # So if we subsequently do replace(J, {g: h}) we end up with an expression:
    # derivative(f(h, h), h)
    # rather than what were were probably thinking of:
    # replace(derivative(f(g, h), g), {g: h})
    #
    # To fix this would require one to expand derivatives early (which
    # is not attractive), or make replace lazy too.
    if has_exact_type(e, CoefficientDerivative):
        # Hack to avoid circular dependencies
        from ufl.algorithms.ad import expand_derivatives

        e = expand_derivatives(e)

    return map_integrand_dags(Replacer(mapping2), e)
